# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: user-admin.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from . import types_pb2 as types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10user-admin.proto\x12\x10\x61\x65rospike.vector\x1a\x1bgoogle/protobuf/empty.proto\x1a\x0btypes.proto\"S\n\x0e\x41\x64\x64UserRequest\x12\x32\n\x0b\x63redentials\x18\x01 \x01(\x0b\x32\x1d.aerospike.vector.Credentials\x12\r\n\x05roles\x18\x02 \x03(\t\"N\n\x18UpdateCredentialsRequest\x12\x32\n\x0b\x63redentials\x18\x01 \x01(\x0b\x32\x1d.aerospike.vector.Credentials\"#\n\x0f\x44ropUserRequest\x12\x10\n\x08username\x18\x01 \x01(\t\"\"\n\x0eGetUserRequest\x12\x10\n\x08username\x18\x01 \x01(\t\"4\n\x11GrantRolesRequest\x12\x10\n\x08username\x18\x01 \x01(\t\x12\r\n\x05roles\x18\x02 \x03(\t\"5\n\x12RevokeRolesRequest\x12\x10\n\x08username\x18\x01 \x01(\t\x12\r\n\x05roles\x18\x02 \x03(\t\":\n\x11ListRolesResponse\x12%\n\x05roles\x18\x01 \x03(\x0b\x32\x16.aerospike.vector.Role\":\n\x11ListUsersResponse\x12%\n\x05users\x18\x01 \x03(\x0b\x32\x16.aerospike.vector.User2\xf8\x04\n\x10UserAdminService\x12\x45\n\x07\x41\x64\x64User\x12 .aerospike.vector.AddUserRequest\x1a\x16.google.protobuf.Empty\"\x00\x12Y\n\x11UpdateCredentials\x12*.aerospike.vector.UpdateCredentialsRequest\x1a\x16.google.protobuf.Empty\"\x00\x12G\n\x08\x44ropUser\x12!.aerospike.vector.DropUserRequest\x1a\x16.google.protobuf.Empty\"\x00\x12\x45\n\x07GetUser\x12 .aerospike.vector.GetUserRequest\x1a\x16.aerospike.vector.User\"\x00\x12J\n\tListUsers\x12\x16.google.protobuf.Empty\x1a#.aerospike.vector.ListUsersResponse\"\x00\x12K\n\nGrantRoles\x12#.aerospike.vector.GrantRolesRequest\x1a\x16.google.protobuf.Empty\"\x00\x12M\n\x0bRevokeRoles\x12$.aerospike.vector.RevokeRolesRequest\x1a\x16.google.protobuf.Empty\"\x00\x12J\n\tListRoles\x12\x16.google.protobuf.Empty\x1a#.aerospike.vector.ListRolesResponse\"\x00\x42\x43\n!com.aerospike.vector.client.protoP\x01Z\x1c\x61\x65rospike.com/vector/protos/b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'user_admin_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n!com.aerospike.vector.client.protoP\001Z\034aerospike.com/vector/protos/'
  _globals['_ADDUSERREQUEST']._serialized_start=80
  _globals['_ADDUSERREQUEST']._serialized_end=163
  _globals['_UPDATECREDENTIALSREQUEST']._serialized_start=165
  _globals['_UPDATECREDENTIALSREQUEST']._serialized_end=243
  _globals['_DROPUSERREQUEST']._serialized_start=245
  _globals['_DROPUSERREQUEST']._serialized_end=280
  _globals['_GETUSERREQUEST']._serialized_start=282
  _globals['_GETUSERREQUEST']._serialized_end=316
  _globals['_GRANTROLESREQUEST']._serialized_start=318
  _globals['_GRANTROLESREQUEST']._serialized_end=370
  _globals['_REVOKEROLESREQUEST']._serialized_start=372
  _globals['_REVOKEROLESREQUEST']._serialized_end=425
  _globals['_LISTROLESRESPONSE']._serialized_start=427
  _globals['_LISTROLESRESPONSE']._serialized_end=485
  _globals['_LISTUSERSRESPONSE']._serialized_start=487
  _globals['_LISTUSERSRESPONSE']._serialized_end=545
  _globals['_USERADMINSERVICE']._serialized_start=548
  _globals['_USERADMINSERVICE']._serialized_end=1180
# @@protoc_insertion_point(module_scope)



from typing import Any

from .. import types
from .proto_generated import types_pb2


def toVectorDbValue(value: Any) -> types_pb2.Value:
    if isinstance(value, str):
        return types_pb2.Value(stringValue=value)
    elif isinstance(value, int):
        return types_pb2.Value(longValue=value)
    elif isinstance(value, float):
        return types_pb2.Value(doubleValue=value)
    elif isinstance(value, (bytes, bytearray)):
        return types_pb2.Value(bytesValue=value)
    elif isinstance(value, bool):
        return types_pb2.Value(booleanValue=value)
    elif isinstance(value, list) and value:
        # TODO: Convert every element correctly to destination type.
        if isinstance(value[0], float):
            return types_pb2.Value(
                vectorValue=types_pb2.Vector(
                    floatData={"value": [float(x) for x in value]}
                )
            )
        elif isinstance(value[0], bool):
            return types_pb2.Value(
                vectorValue=types_pb2.Vector(
                    boolData={"value": [True if x else False for x in value]}
                )
            )
        else:
            return types_pb2.Value(
                listValue=types_pb2.List(entries=[toVectorDbValue(x) for x in value])
            )
    elif isinstance(value, dict):
        d = types_pb2.Value(
            mapValue=types_pb2.Map(
                entries=[
                    types_pb2.MapEntry(key=toMapKey(k), value=toVectorDbValue(v))
                    for k, v in value.items()
                ]
            )
        )
        return d
    else:
        raise Exception("Invalid type " + str(type(value)))


def toMapKey(value):
    if isinstance(value, str):
        return types_pb2.MapKey(stringValue=value)
    elif isinstance(value, int):
        return types_pb2.MapKey(longValue=value)
    elif isinstance(value, (bytes, bytearray)):
        return types_pb2.MapKey(bytesValue=value)
    elif isinstance(value, float):
        return types_pb2.MapKey(doubleValue=value)
    else:
        raise Exception("Invalid map key type " + str(type(value)))


def fromVectorDbKey(key: types_pb2.Key) -> types.Key:
    keyValue = None
    if key.HasField("stringValue"):
        keyValue = key.stringValue
    elif key.HasField("intValue"):
        keyValue = key.intValue
    elif key.HasField("longValue"):
        keyValue = key.longValue
    elif key.HasField("bytesValue"):
        keyValue = key.bytesValue

    return types.Key(namespace=key.namespace, set=key.set, key=keyValue)


def fromVectorDbRecord(record: types_pb2.Record) -> dict[str, Any]:
    fields = {}
    for field in record.fields:
        fields[field.name] = fromVectorDbValue(field.value)

    return fields


def fromVectorDbNeighbor(input: types_pb2.Neighbor) -> types.Neighbor:
    return types.Neighbor(
        key=fromVectorDbKey(input.key),
        fields=fromVectorDbRecord(input.record),
        distance=input.distance,
    )


def fromIndexDefintion(input: types_pb2.IndexDefinition) -> types.IndexDefinition:
    return types.IndexDefinition(
        id=types.IndexId(
            namespace=input.id.namespace,
            name=input.id.name,
        ),
        dimensions=input.dimensions,
        vector_distance_metric=input.vectorDistanceMetric,
        field=input.field,
        sets=input.setFilter,
        hnsw_params=types.HnswParams(
            m=input.hnswParams.m,
            ef_construction=input.hnswParams.efConstruction,
            ef=input.hnswParams.ef,
            batching_params=types.HnswBatchingParams(
                max_records=input.hnswParams.batchingParams.maxRecords,
                interval=input.hnswParams.batchingParams.interval,
            ),
            caching_params=types.HnswCachingParams(
                max_entries=input.hnswParams.cachingParams.maxEntries,
                expiry=input.hnswParams.cachingParams.expiry,
            ),
            healer_params=types.HnswHealerParams(
               max_scan_rate_per_node=input.hnswParams.healerParams.maxScanRatePerNode,
               max_scan_page_size=input.hnswParams.healerParams.maxScanPageSize,
               re_index_percent=input.hnswParams.healerParams.reindexPercent,
               schedule_delay=input.hnswParams.healerParams.scheduleDelay,
               parallelism=input.hnswParams.healerParams.parallelism
            ),
            merge_params=types.HnswIndexMergeParams(
               parallelism=input.hnswParams.mergeParams.parallelism
            )
        ),
        index_labels=input.labels,
        storage=types.IndexStorage(
            namespace=input.storage.namespace, set_name=input.storage.set
        ),
    )


def fromVectorDbValue(input: types_pb2.Value) -> Any:
    if input.HasField("stringValue"):
        return input.stringValue
    elif input.HasField("intValue"):
        return input.intValue
    elif input.HasField("longValue"):
        return input.longValue
    elif input.HasField("bytesValue"):
        return input.bytesValue
    elif input.HasField("mapValue"):
        dict = {}
        for entry in input.mapValue.entries:
            k = fromVectorDbValue(entry.key)
            v = fromVectorDbValue(entry.value)
            dict[k] = v
        return dict
    elif input.HasField("listValue"):
        return [fromVectorDbValue(v) for v in input.listValue.entries]
    elif input.HasField("vectorValue"):
        vector = input.vectorValue
        if vector.HasField("floatData"):
            return [v for v in vector.floatData.value]
        if vector.HasField("boolData"):
            return [v for v in vector.boolData.value]

    return None



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: index.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from . import types_pb2 as types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0bindex.proto\x12\x10\x61\x65rospike.vector\x1a\x1bgoogle/protobuf/empty.proto\x1a\x0btypes.proto\"2\n\x13IndexStatusResponse\x12\x1b\n\x13unmergedRecordCount\x18\x02 \x01(\x03\"_\n\x18GcInvalidVerticesRequest\x12*\n\x07indexId\x18\x01 \x01(\x0b\x32\x19.aerospike.vector.IndexId\x12\x17\n\x0f\x63utoffTimestamp\x18\x02 \x01(\x03\"\xf9\x01\n\x12IndexUpdateRequest\x12*\n\x07indexId\x18\x01 \x01(\x0b\x32\x19.aerospike.vector.IndexId\x12@\n\x06labels\x18\x02 \x03(\x0b\x32\x30.aerospike.vector.IndexUpdateRequest.LabelsEntry\x12<\n\x0fhnswIndexUpdate\x18\x03 \x01(\x0b\x32!.aerospike.vector.HnswIndexUpdateH\x00\x1a-\n\x0bLabelsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\x08\n\x06update2\x98\x04\n\x0cIndexService\x12\x45\n\x06\x43reate\x12!.aerospike.vector.IndexDefinition\x1a\x16.google.protobuf.Empty\"\x00\x12H\n\x06Update\x12$.aerospike.vector.IndexUpdateRequest\x1a\x16.google.protobuf.Empty\"\x00\x12;\n\x04\x44rop\x12\x19.aerospike.vector.IndexId\x1a\x16.google.protobuf.Empty\"\x00\x12G\n\x04List\x12\x16.google.protobuf.Empty\x1a%.aerospike.vector.IndexDefinitionList\"\x00\x12\x45\n\x03Get\x12\x19.aerospike.vector.IndexId\x1a!.aerospike.vector.IndexDefinition\"\x00\x12O\n\tGetStatus\x12\x19.aerospike.vector.IndexId\x1a%.aerospike.vector.IndexStatusResponse\"\x00\x12Y\n\x11GcInvalidVertices\x12*.aerospike.vector.GcInvalidVerticesRequest\x1a\x16.google.protobuf.Empty\"\x00\x42\x43\n!com.aerospike.vector.client.protoP\x01Z\x1c\x61\x65rospike.com/vector/protos/b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'index_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n!com.aerospike.vector.client.protoP\001Z\034aerospike.com/vector/protos/'
  _globals['_INDEXUPDATEREQUEST_LABELSENTRY']._loaded_options = None
  _globals['_INDEXUPDATEREQUEST_LABELSENTRY']._serialized_options = b'8\001'
  _globals['_INDEXSTATUSRESPONSE']._serialized_start=75
  _globals['_INDEXSTATUSRESPONSE']._serialized_end=125
  _globals['_GCINVALIDVERTICESREQUEST']._serialized_start=127
  _globals['_GCINVALIDVERTICESREQUEST']._serialized_end=222
  _globals['_INDEXUPDATEREQUEST']._serialized_start=225
  _globals['_INDEXUPDATEREQUEST']._serialized_end=474
  _globals['_INDEXUPDATEREQUEST_LABELSENTRY']._serialized_start=419
  _globals['_INDEXUPDATEREQUEST_LABELSENTRY']._serialized_end=464
  _globals['_INDEXSERVICE']._serialized_start=477
  _globals['_INDEXSERVICE']._serialized_end=1013
# @@protoc_insertion_point(module_scope)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from . import transact_pb2 as transact__pb2
from . import types_pb2 as types__pb2

GRPC_GENERATED_VERSION = '1.64.1'
GRPC_VERSION = grpc.__version__
EXPECTED_ERROR_RELEASE = '1.65.0'
SCHEDULED_RELEASE_DATE = 'June 25, 2024'
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    warnings.warn(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in transact_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
        + f' This warning will become an error in {EXPECTED_ERROR_RELEASE},'
        + f' scheduled for release on {SCHEDULED_RELEASE_DATE}.',
        RuntimeWarning
    )


class TransactServiceStub(object):
    """Record transaction services.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Put = channel.unary_unary(
                '/aerospike.vector.TransactService/Put',
                request_serializer=transact__pb2.PutRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.Get = channel.unary_unary(
                '/aerospike.vector.TransactService/Get',
                request_serializer=transact__pb2.GetRequest.SerializeToString,
                response_deserializer=types__pb2.Record.FromString,
                _registered_method=True)
        self.Delete = channel.unary_unary(
                '/aerospike.vector.TransactService/Delete',
                request_serializer=transact__pb2.DeleteRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.Exists = channel.unary_unary(
                '/aerospike.vector.TransactService/Exists',
                request_serializer=transact__pb2.ExistsRequest.SerializeToString,
                response_deserializer=types__pb2.Boolean.FromString,
                _registered_method=True)
        self.IsIndexed = channel.unary_unary(
                '/aerospike.vector.TransactService/IsIndexed',
                request_serializer=transact__pb2.IsIndexedRequest.SerializeToString,
                response_deserializer=types__pb2.Boolean.FromString,
                _registered_method=True)
        self.VectorSearch = channel.unary_stream(
                '/aerospike.vector.TransactService/VectorSearch',
                request_serializer=transact__pb2.VectorSearchRequest.SerializeToString,
                response_deserializer=types__pb2.Neighbor.FromString,
                _registered_method=True)


class TransactServiceServicer(object):
    """Record transaction services.
    """

    def Put(self, request, context):
        """Update/insert records.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Get(self, request, context):
        """Get a record.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Delete(self, request, context):
        """Delete a record.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Exists(self, request, context):
        """Check if a record exists.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def IsIndexed(self, request, context):
        """Check is a record is indexed.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def VectorSearch(self, request, context):
        """Perform a vector nearest neighbor search.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_TransactServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Put': grpc.unary_unary_rpc_method_handler(
                    servicer.Put,
                    request_deserializer=transact__pb2.PutRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'Get': grpc.unary_unary_rpc_method_handler(
                    servicer.Get,
                    request_deserializer=transact__pb2.GetRequest.FromString,
                    response_serializer=types__pb2.Record.SerializeToString,
            ),
            'Delete': grpc.unary_unary_rpc_method_handler(
                    servicer.Delete,
                    request_deserializer=transact__pb2.DeleteRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'Exists': grpc.unary_unary_rpc_method_handler(
                    servicer.Exists,
                    request_deserializer=transact__pb2.ExistsRequest.FromString,
                    response_serializer=types__pb2.Boolean.SerializeToString,
            ),
            'IsIndexed': grpc.unary_unary_rpc_method_handler(
                    servicer.IsIndexed,
                    request_deserializer=transact__pb2.IsIndexedRequest.FromString,
                    response_serializer=types__pb2.Boolean.SerializeToString,
            ),
            'VectorSearch': grpc.unary_stream_rpc_method_handler(
                    servicer.VectorSearch,
                    request_deserializer=transact__pb2.VectorSearchRequest.FromString,
                    response_serializer=types__pb2.Neighbor.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'aerospike.vector.TransactService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('aerospike.vector.TransactService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class TransactService(object):
    """Record transaction services.
    """

    @staticmethod
    def Put(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/aerospike.vector.TransactService/Put',
            transact__pb2.PutRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Get(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/aerospike.vector.TransactService/Get',
            transact__pb2.GetRequest.SerializeToString,
            types__pb2.Record.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Delete(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/aerospike.vector.TransactService/Delete',
            transact__pb2.DeleteRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Exists(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/aerospike.vector.TransactService/Exists',
            transact__pb2.ExistsRequest.SerializeToString,
            types__pb2.Boolean.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def IsIndexed(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/aerospike.vector.TransactService/IsIndexed',
            transact__pb2.IsIndexedRequest.SerializeToString,
            types__pb2.Boolean.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def VectorSearch(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/aerospike.vector.TransactService/VectorSearch',
            transact__pb2.VectorSearchRequest.SerializeToString,
            types__pb2.Neighbor.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



import asyncio
import logging
from typing import Any, Optional, Union
import time

import google.protobuf.empty_pb2
from google.protobuf.json_format import MessageToDict
import grpc

from . import helpers
from .proto_generated import index_pb2_grpc, user_admin_pb2_grpc
from .proto_generated import types_pb2, user_admin_pb2
from .. import types
from . import conversions

logger = logging.getLogger(__name__)

empty = google.protobuf.empty_pb2.Empty()


class BaseClient(object):

    def _prepare_seeds(self, seeds) -> None:
        return helpers._prepare_seeds(seeds)

    def _prepare_index_create(
        self,
        namespace,
        name,
        vector_field,
        dimensions,
        vector_distance_metric,
        sets,
        index_params,
        index_labels,
        index_storage,
        timeout,
        logger,
    ) -> None:

        logger.debug(
            "Creating index: namespace=%s, name=%s, vector_field=%s, dimensions=%d, vector_distance_metric=%s, "
            "sets=%s, index_params=%s, index_labels=%s, index_storage=%s, timeout=%s",
            namespace,
            name,
            vector_field,
            dimensions,
            vector_distance_metric,
            sets,
            index_params,
            index_labels,
            index_storage,
            timeout,
        )

        kwargs = {}
        if timeout is not None:
            kwargs["timeout"] = timeout

        if sets and not sets.strip():
            sets = None
        if index_params != None:
            index_params = index_params._to_pb2()
        if index_storage != None:
            index_storage = index_storage._to_pb2()

        id = self._get_index_id(namespace, name)
        vector_distance_metric = vector_distance_metric.value

        index_stub = self._get_index_stub()

        index_create_request = types_pb2.IndexDefinition(
            id=id,
            vectorDistanceMetric=vector_distance_metric,
            setFilter=sets,
            hnswParams=index_params,
            field=vector_field,
            dimensions=dimensions,
            labels=index_labels,
            storage=index_storage,
        )
        return (index_stub, index_create_request, kwargs)

    def _prepare_index_drop(self, namespace, name, timeout, logger) -> None:

        logger.debug(
            "Dropping index: namespace=%s, name=%s, timeout=%s",
            namespace,
            name,
            timeout,
        )

        kwargs = {}
        if timeout is not None:
            kwargs["timeout"] = timeout

        index_stub = self._get_index_stub()
        index_drop_request = self._get_index_id(namespace, name)

        return (index_stub, index_drop_request, kwargs)

    def _prepare_index_list(self, timeout, logger) -> None:

        logger.debug("Getting index list: timeout=%s", timeout)

        kwargs = {}
        if timeout is not None:
            kwargs["timeout"] = timeout

        index_stub = self._get_index_stub()
        index_list_request = empty

        return (index_stub, index_list_request, kwargs)

    def _prepare_index_get(self, namespace, name, timeout, logger) -> None:

        logger.debug(
            "Getting index information: namespace=%s, name=%s, timeout=%s",
            namespace,
            name,
            timeout,
        )

        kwargs = {}
        if timeout is not None:
            kwargs["timeout"] = timeout

        index_stub = self._get_index_stub()
        index_get_request = self._get_index_id(namespace, name)

        return (index_stub, index_get_request, kwargs)

    def _prepare_index_get_status(self, namespace, name, timeout, logger) -> None:

        logger.debug(
            "Getting index status: namespace=%s, name=%s, timeout=%s",
            namespace,
            name,
            timeout,
        )

        kwargs = {}
        if timeout is not None:
            kwargs["timeout"] = timeout

        index_stub = self._get_index_stub()
        index_get_status_request = self._get_index_id(namespace, name)

        return (index_stub, index_get_status_request, kwargs)

    def _prepare_add_user(self, username, password, roles, timeout, logger) -> None:
        logger.debug(
            "Getting index status: username=%s, password=%s, roles=%s, timeout=%s",
            username,
            password,
            roles,
            timeout,
        )

        kwargs = {}
        if timeout is not None:
            kwargs["timeout"] = timeout

        user_admin_stub = self._get_user_admin_stub()
        credentials = helpers._get_credentials(username, password)
        add_user_request = user_admin_pb2.AddUserRequest(
            credentials=credentials, roles=roles
        )

        return (user_admin_stub, add_user_request, kwargs)

    def _prepare_update_credentials(self, username, password, timeout, logger) -> None:
        logger.debug(
            "Getting index status: username=%s, password=%s, timeout=%s",
            username,
            password,
            timeout,
        )

        kwargs = {}
        if timeout is not None:
            kwargs["timeout"] = timeout

        user_admin_stub = self._get_user_admin_stub()
        credentials = helpers._get_credentials(username, password)
        update_user_request = user_admin_pb2.UpdateCredentialsRequest(
            credentials=credentials
        )

        return (user_admin_stub, update_user_request, kwargs)

    def _prepare_drop_user(self, username, timeout, logger) -> None:
        logger.debug("Getting index status: username=%s, timeout=%s", username, timeout)

        kwargs = {}
        if timeout is not None:
            kwargs["timeout"] = timeout

        user_admin_stub = self._get_user_admin_stub()
        drop_user_request = user_admin_pb2.DropUserRequest(username=username)

        return (user_admin_stub, drop_user_request, kwargs)

    def _prepare_get_user(self, username, timeout, logger) -> None:
        logger.debug("Getting index status: username=%s, timeout=%s", username, timeout)

        kwargs = {}
        if timeout is not None:
            kwargs["timeout"] = timeout

        user_admin_stub = self._get_user_admin_stub()
        get_user_request = user_admin_pb2.GetUserRequest(username=username)

        return (user_admin_stub, get_user_request, kwargs)

    def _prepare_list_users(self, timeout, logger) -> None:
        logger.debug("Getting index status")

        kwargs = {}
        if timeout is not None:
            kwargs["timeout"] = timeout

        user_admin_stub = self._get_user_admin_stub()
        list_users_request = empty

        return (user_admin_stub, list_users_request, kwargs)

    def _prepare_grant_roles(self, username, roles, timeout, logger) -> None:
        logger.debug(
            "Getting index status: username=%s, roles=%s, timeout=%s",
            username,
            roles,
            timeout,
        )

        kwargs = {}
        if timeout is not None:
            kwargs["timeout"] = timeout

        user_admin_stub = self._get_user_admin_stub()
        grant_roles_request = user_admin_pb2.GrantRolesRequest(
            username=username, roles=roles
        )

        return (user_admin_stub, grant_roles_request, kwargs)

    def _prepare_revoke_roles(self, username, roles, timeout, logger) -> None:
        logger.debug(
            "Getting index status: username=%s, roles=%s, timeout=%s",
            username,
            roles,
            timeout,
        )

        kwargs = {}
        if timeout is not None:
            kwargs["timeout"] = timeout

        user_admin_stub = self._get_user_admin_stub()
        revoke_roles_request = user_admin_pb2.RevokeRolesRequest(
            username=username, roles=roles
        )

        return (user_admin_stub, revoke_roles_request, kwargs)

    def _prepare_list_roles(self, timeout, logger) -> None:
        logger.debug("Getting index status: timeout=%s", timeout)

        kwargs = {}
        if timeout is not None:
            kwargs["timeout"] = timeout

        user_admin_stub = self._get_user_admin_stub()
        list_roles_request = empty

        return (user_admin_stub, list_roles_request, kwargs)

    def _respond_index_list(self, response) -> None:
        response_list = []
        for index in response.indices:
            response_list.append(conversions.fromIndexDefintion(index))
        return response_list

    def _respond_index_get(self, response) -> None:

        return conversions.fromIndexDefintion(response)

    def _respond_get_user(self, response) -> None:

        return types.User(username=response.username, roles=list(response.roles))

    def _respond_list_users(self, response) -> None:
        user_list = []
        for user in response.users:
            user_list.append(types.User(username=user.username, roles=list(user.roles)))
        return user_list

    def _respond_list_roles(self, response) -> None:
        role_list = []
        for role in response.roles:
            role_list.append(types.Role(id=role.id))
        return role_list

    def _respond_index_get_status(self, response) -> None:
        return response.unmergedRecordCount

    def _get_index_stub(self):
        return index_pb2_grpc.IndexServiceStub(self._channel_provider.get_channel())

    def _get_user_admin_stub(self):
        return user_admin_pb2_grpc.UserAdminServiceStub(
            self._channel_provider.get_channel()
        )

    def _get_index_id(self, namespace, name):
        return types_pb2.IndexId(namespace=namespace, name=name)

    def _get_add_user_request(self, namespace, name):
        return user_admin_pb2.AddUserRequest(namespace=namespace, name=name)

    def _prepare_wait_for_index_waiting(self, namespace, name, wait_interval):
        return helpers._prepare_wait_for_index_waiting(
            self, namespace, name, wait_interval
        )

    def _check_timeout(self, start_time, timeout):
        if start_time + timeout < time.monotonic():
            raise AVSClientError(message="timed-out waiting for index creation")



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from . import index_pb2 as index__pb2
from . import types_pb2 as types__pb2

GRPC_GENERATED_VERSION = '1.64.1'
GRPC_VERSION = grpc.__version__
EXPECTED_ERROR_RELEASE = '1.65.0'
SCHEDULED_RELEASE_DATE = 'June 25, 2024'
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    warnings.warn(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in index_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
        + f' This warning will become an error in {EXPECTED_ERROR_RELEASE},'
        + f' scheduled for release on {SCHEDULED_RELEASE_DATE}.',
        RuntimeWarning
    )


class IndexServiceStub(object):
    """Service to manage indices.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Create = channel.unary_unary(
                '/aerospike.vector.IndexService/Create',
                request_serializer=types__pb2.IndexDefinition.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.Update = channel.unary_unary(
                '/aerospike.vector.IndexService/Update',
                request_serializer=index__pb2.IndexUpdateRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.Drop = channel.unary_unary(
                '/aerospike.vector.IndexService/Drop',
                request_serializer=types__pb2.IndexId.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)
        self.List = channel.unary_unary(
                '/aerospike.vector.IndexService/List',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=types__pb2.IndexDefinitionList.FromString,
                _registered_method=True)
        self.Get = channel.unary_unary(
                '/aerospike.vector.IndexService/Get',
                request_serializer=types__pb2.IndexId.SerializeToString,
                response_deserializer=types__pb2.IndexDefinition.FromString,
                _registered_method=True)
        self.GetStatus = channel.unary_unary(
                '/aerospike.vector.IndexService/GetStatus',
                request_serializer=types__pb2.IndexId.SerializeToString,
                response_deserializer=index__pb2.IndexStatusResponse.FromString,
                _registered_method=True)
        self.GcInvalidVertices = channel.unary_unary(
                '/aerospike.vector.IndexService/GcInvalidVertices',
                request_serializer=index__pb2.GcInvalidVerticesRequest.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                _registered_method=True)


class IndexServiceServicer(object):
    """Service to manage indices.
    """

    def Create(self, request, context):
        """Create an index.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Update(self, request, context):
        """Create an index.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Drop(self, request, context):
        """Drop an index.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def List(self, request, context):
        """List available indices.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Get(self, request, context):
        """Get the index definition.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetStatus(self, request, context):
        """Query status of an index.
        NOTE: API is subject to change.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GcInvalidVertices(self, request, context):
        """Garbage collect vertices identified as invalid before cutoff timestamp.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_IndexServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Create': grpc.unary_unary_rpc_method_handler(
                    servicer.Create,
                    request_deserializer=types__pb2.IndexDefinition.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'Update': grpc.unary_unary_rpc_method_handler(
                    servicer.Update,
                    request_deserializer=index__pb2.IndexUpdateRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'Drop': grpc.unary_unary_rpc_method_handler(
                    servicer.Drop,
                    request_deserializer=types__pb2.IndexId.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'List': grpc.unary_unary_rpc_method_handler(
                    servicer.List,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=types__pb2.IndexDefinitionList.SerializeToString,
            ),
            'Get': grpc.unary_unary_rpc_method_handler(
                    servicer.Get,
                    request_deserializer=types__pb2.IndexId.FromString,
                    response_serializer=types__pb2.IndexDefinition.SerializeToString,
            ),
            'GetStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.GetStatus,
                    request_deserializer=types__pb2.IndexId.FromString,
                    response_serializer=index__pb2.IndexStatusResponse.SerializeToString,
            ),
            'GcInvalidVertices': grpc.unary_unary_rpc_method_handler(
                    servicer.GcInvalidVertices,
                    request_deserializer=index__pb2.GcInvalidVerticesRequest.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'aerospike.vector.IndexService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('aerospike.vector.IndexService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class IndexService(object):
    """Service to manage indices.
    """

    @staticmethod
    def Create(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/aerospike.vector.IndexService/Create',
            types__pb2.IndexDefinition.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Update(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/aerospike.vector.IndexService/Update',
            index__pb2.IndexUpdateRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Drop(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/aerospike.vector.IndexService/Drop',
            types__pb2.IndexId.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def List(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/aerospike.vector.IndexService/List',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            types__pb2.IndexDefinitionList.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Get(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/aerospike.vector.IndexService/Get',
            types__pb2.IndexId.SerializeToString,
            types__pb2.IndexDefinition.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/aerospike.vector.IndexService/GetStatus',
            types__pb2.IndexId.SerializeToString,
            index__pb2.IndexStatusResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GcInvalidVertices(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/aerospike.vector.IndexService/GcInvalidVertices',
            index__pb2.GcInvalidVerticesRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



import time
from .. import types
from .proto_generated import types_pb2
from .proto_generated import index_pb2_grpc


def _prepare_seeds(seeds) -> None:

    if not seeds:
        raise types.AVSClientError(message="at least one seed host needed")

    if isinstance(seeds, types.HostPort):
        seeds = (seeds,)

    return seeds


def _prepare_wait_for_index_waiting(client, namespace, name, wait_interval):

    unmerged_record_initialized = False
    start_time = time.monotonic()
    consecutive_index_validations = 0

    index_stub = index_pb2_grpc.IndexServiceStub(client._channel_provider.get_channel())
    index_wait_request = types_pb2.IndexId(namespace=namespace, name=name)
    return (
        index_stub,
        wait_interval,
        start_time,
        unmerged_record_initialized,
        consecutive_index_validations,
        index_wait_request,
    )


def _get_credentials(username, password):
    if not username:
        return None
    return types_pb2.Credentials(
        username=username,
        passwordCredentials=types_pb2.PasswordCredentials(password=password),
    )



import enum
from typing import Any, Optional

from .shared.proto_generated import types_pb2


class HostPort(object):
    """
    represents host, port and TLS usage information.
    Used primarily when intializing client.

    :param host: The host address.
    :type host: str

    :param port: The port number.
    :type port: int

    """

    def __init__(self, *, host: str, port: int) -> None:
        self.host = host
        self.port = port


class Key(object):
    """
    Represents a record key.
    Used in RecordWithKey.

    :param namespace (str): The namespace for the key.
    :type namespace: str

    :param set: (optional[str]): The set for the key.
    :type set: optional[str]

    :param key: (Any): The key itself.
    :type key: Union[int, str, bytes, bytearray, np.ndarray, np.generic]

    """

    def __init__(self, *, namespace: str, set: str, key: Any) -> None:
        self.namespace = namespace
        self.set = set
        self.key = key

    def __str__(self):
        """
        Returns a string representation of the key.
        """
        return f"Key: namespace='{self.namespace}', set='{self.set}', key={self.key}"


class RecordWithKey(object):
    """
    Represents a record, including a key and fields.
    Return value for VectorDbClient.get.

    :param key: (Key): The key of the record.
    :type key: Key

    :param fields: : The fields associated with the record.
    :type fields: dict[str, Any]
    """

    def __init__(self, *, key: Key, fields: dict[str, Any]) -> None:
        self.key = key
        self.fields = fields

    def __str__(self):
        """
        Returns a string representation of the record, including a key and fields.
        """
        fields_info = ""
        for key, value in self.fields.items():
            if isinstance(value, list):
                if len(value) > 4:
                    value_str = (
                        "[\n"
                        + ",\n".join("\t\t\t{}".format(val) for val in value[:3])
                        + ",\n\t\t\t...\n\t\t]"
                    )
                else:
                    value_str = str(value)
            else:
                value_str = str(value)
            fields_info += "\n\t\t{}: {}".format(key, value_str)
        return "{{\n\t{},\n\tfields: {{\n{}\n\t}}\n}}".format(self.key, fields_info)


class Neighbor(object):
    """
    Represents a neighboring record in the context of approximate nearest neighbor search.

    This class represents a neighboring record in relation to a query record. It includes information such as the key, fields,
        and distance from the query record.


    :param key: The Key instance identifying the neighboring record.
    :type distance: Key

    :param fields: A dictionary representing fields associated with the neighboring record.
    :type distance: dict[str, Any]

    :param distance: The distance between the neighboring record and the query record, calculated based on the chosen
        VectorDistanceMetric.
    :type distance: float

    Notes:
        - The distance metric used to calculate the distance between records is determined by the chosen VectorDistanceMetric.
        - The neighbor's distance indicates how similar or dissimilar it is to the query record based on the chosen distance metric.
        - A smaller distance typically implies greater similarity between records.

    """

    def __init__(self, *, key: Key, fields: dict[str, Any], distance: float) -> None:
        self.key = key
        self.fields = fields
        self.distance = distance

    def __str__(self):
        """
        Returns a string representation of the neighboring record.
        """
        fields_info = ""
        for key, value in self.fields.items():
            if isinstance(value, list):
                if len(value) > 4:
                    value_str = (
                        "[\n"
                        + ",\n".join("\t\t\t{}".format(val) for val in value[:3])
                        + ",\n\t\t\t...\n\t\t]"
                    )
                else:
                    value_str = str(value)
            else:
                value_str = str(value)
            fields_info += "\n\t\t{}: {}".format(key, value_str)
        return "{{\n\t{},\n\tdistance: {},\n\tfields: {{\n{}\n\t}}\n}}".format(
            self.key, self.distance, fields_info
        )


class VectorDistanceMetric(enum.Enum):
    """
    Enumeration of vector distance metrics used for comparing vectors.

    This enumeration defines various metrics for calculating the distance or similarity between vectors:

    - **SQUARED_EUCLIDEAN**: Represents the squared Euclidean distance metric.
    - **COSINE**: Represents the cosine similarity metric.
    - **DOT_PRODUCT**: Represents the dot product similarity metric.
    - **MANHATTAN**: Represents the Manhattan distance (L1 norm) metric.
    - **HAMMING**: Represents the Hamming distance metric.

    Each metric provides a different method for comparing vectors, affecting how distances and similarities are computed in vector-based operations.
    """

    SQUARED_EUCLIDEAN: types_pb2.VectorDistanceMetric = (
        types_pb2.VectorDistanceMetric.SQUARED_EUCLIDEAN
    )
    COSINE: types_pb2.VectorDistanceMetric = types_pb2.VectorDistanceMetric.COSINE
    DOT_PRODUCT: types_pb2.VectorDistanceMetric = (
        types_pb2.VectorDistanceMetric.DOT_PRODUCT
    )
    MANHATTAN: types_pb2.VectorDistanceMetric = types_pb2.VectorDistanceMetric.MANHATTAN
    HAMMING: types_pb2.VectorDistanceMetric = types_pb2.VectorDistanceMetric.HAMMING


class Role(object):
    """
    AVS Role Object used in role-based authentication.

    :param id: Unique id to identify role
    :type id: str

    """

    def __init__(
        self,
        *,
        id: str,
    ) -> None:
        self.id = id

    def __repr__(self) -> str:
        return f"Role(id={self.id!r})"

    def __str__(self) -> str:
        return f"Role {{\n  id: {self.id}\n}}"

    def __getitem__(self, key):
        key = str(key)  # Ensure key is a string
        if not hasattr(self, key):
            raise AttributeError(f"'Role' object has no attribute '{key}'")
        return getattr(self, key)

    def __setitem__(self, key, value):
        setattr(self, key, value)


class User(object):
    """
    AVS User Object used in role-based authentication.

    :param username: Username associated with user.
    :type username: str

    :param roles: roles associated with user.
    :type roles: list[str]

    """

    def __init__(
        self,
        *,
        username: str,
        roles: Optional[list[Role]],
    ) -> None:
        self.username = username
        self.roles = roles


class HnswBatchingParams(object):
    """
    Parameters for configuring batching behaviour for batch based index update.

    :param max_records: Maximum number of records to fit in a batch. Defaults to server default..
    :param interval: The maximum amount of time in milliseconds to wait before finalizing a batch. Defaults to server default..
    """

    def __init__(
        self,
        *,
        max_records: Optional[int] = None,
        interval: Optional[int] = None,
    ) -> None:
        self.max_records = max_records
        self.interval = interval

    def _to_pb2(self):
        params = types_pb2.HnswBatchingParams()
        if self.max_records:
            params.maxRecords = self.max_records
        if self.interval:
            params.interval = self.interval
        return params

    def __repr__(self) -> str:
        return (
            f"batchingParams(max_records={self.max_records}, "
            f"interval={self.interval})"
        )

    def __str__(self) -> str:
        return (
            f"batchingParams {{\n"
            f"  maxRecords: {self.max_records}\n"
            f"  interval: {self.interval}\n"
            f"}}"
        )

    def __getitem__(self, key):
        key = str(key)  # Ensure key is a string
        if not hasattr(self, key):
            raise AttributeError(
                f"'HnswBatchingParams' object has no attribute '{key}'"
            )
        return getattr(self, key)

    def __setitem__(self, key, value):
        return setattr(self, key, value)


class HnswHealerParams(object):
    """
    Parameters to configure the HNSW healer

    :param max_scan_rate_per_node: Maximum allowed record scan rate per vector db node.  Default is the global healer config, which is configured in the AVS Server.
    :type max_scan_rate_per_node: Optional[int]

    :param max_scan_page_size: Maximum number of records in a single scanned page.
        Default is the global healer config, which is configured in the AVS Server.
    :type max_scan_page_size: Optional[int]

    :param re_index_percent: Percentage of good records randomly selected for reindexing in a healer cycle.
        Default is the global healer config, which is configured in the AVS Server.
    :type re_index_percent: Optional[int]

    :param schedule_delay: The time delay, in milliseconds, between the termination of a healer run and the commencement
        of the next one for an index. It only guarantees that the next index healer run for an index will not be scheduled before
        this time delay. Default is the global healer config, which is configured in the AVS Server.
    :type schedule_delay: Optional[int]

    :param parallelism: Maximum number of records to heal in parallel.
        Default is the global healer config, which is configured in the AVS Server.
    :type parallelism: Optional[int]
    """

    def __init__(
        self,
        *,
        max_scan_rate_per_node: Optional[int] = None,
        max_scan_page_size: Optional[int] = None,
        re_index_percent: Optional[int] = None,
        schedule_delay: Optional[int] = None,
        parallelism: Optional[int] = None,
    ) -> None:
        self.max_scan_rate_per_node = max_scan_rate_per_node
        self.max_scan_page_size = max_scan_page_size
        self.re_index_percent = re_index_percent
        self.schedule_delay = schedule_delay
        self.parallelism = parallelism

    def _to_pb2(self):
        params = types_pb2.HnswHealerParams()
        if self.max_scan_rate_per_node:
            params.maxScanRatePerNode = self.max_scan_rate_per_node

        if self.max_scan_page_size:

            params.maxScanPageSize = self.max_scan_page_size

        if self.re_index_percent:

            params.reindexPercent = self.re_index_percent

        if self.schedule_delay:

            params.scheduleDelay = self.schedule_delay

        if self.parallelism:

            params.parallelism = self.parallelism

        return params


class HnswCachingParams(object):
    """
    Parameters to configure the HNSW index cache

    :param max_entries: maximum number of entries to cache.  Default is the global cache config, which is configured in the AVS Server.
    :type max_entries: Optional[int]

    :param expiry: Cache entries will expire after this time in millseconds has expired after the entry was add to the cache.
        Default is the global cache config, which is configured in the AVS Server.
    :type expiry: Optional[int]

    """

    def __init__(
        self, *, max_entries: Optional[int] = None, expiry: Optional[int] = None
    ) -> None:
        self.max_entries = max_entries
        self.expiry = expiry

    def _to_pb2(self):
        params = types_pb2.HnswCachingParams()
        if self.max_entries:
            params.maxEntries = self.max_entries
        if self.expiry:
            params.expiry = self.expiry
        return params


class HnswIndexMergeParams(object):
    """
    Parameters to configure the HNSW index merge behavior.

    :param parallelism: The number of vectors merged in parallel from a batch index to main index.
        Default is the global healer config, which is configured in the AVS Server.
    :type parallelism: Optional[int]
    """

    def __init__(self, *, parallelism: Optional[int] = None) -> None:
        self.parallelism = parallelism

    def _to_pb2(self):
        params = types_pb2.HnswIndexMergeParams()
        if self.parallelism:
            params.parallelism = self.parallelism
        return params


class HnswParams(object):
    """
    Parameters for the Hierarchical Navigable Small World (HNSW) algorithm, used for approximate nearest neighbor search.

    :param m: The number of bi-directional links created per level during construction. Larger 'm' values lead to higher recall but slower construction. Defaults to 16.
    :type m: Optional[int]

    :param ef_construction: The size of the dynamic list for the nearest neighbors (candidates) during the index construction. Larger 'ef_construction' values lead to higher recall but slower construction. Defaults to 100.
    :type ef_construction: Optional[int]

    :param ef: The size of the dynamic list for the nearest neighbors (candidates) during the search phase. Larger 'ef' values lead to higher recall but slower search. Defaults to 100.
    :type ef: Optional[int]

    :param batching_params: Parameters related to configuring batch processing, such as the maximum number of records per batch and batching interval. Defaults to HnswBatchingParams().
    :type batching_params: Optional[HnswBatchingParams]
    """

    def __init__(
        self,
        *,
        m: Optional[int] = None,
        ef_construction: Optional[int] = None,
        ef: Optional[int] = None,
        batching_params: Optional[HnswBatchingParams] = HnswBatchingParams(),
        max_mem_queue_size: Optional[int] = None,
        caching_params: Optional[HnswCachingParams] = HnswCachingParams(),
        healer_params: Optional[HnswHealerParams] = HnswHealerParams(),
        merge_params: Optional[HnswIndexMergeParams] = HnswIndexMergeParams(),
    ) -> None:
        self.m = m
        self.ef_construction = ef_construction
        self.ef = ef
        self.batching_params = batching_params
        self.max_mem_queue_size = max_mem_queue_size
        self.caching_params = caching_params
        self.healer_params = healer_params
        self.merge_params = merge_params

    def _to_pb2(self):
        params = types_pb2.HnswParams()
        if self.m:
            params.m = self.m

        if self.ef_construction:
            params.efConstruction = self.ef_construction

        if self.ef:
            params.ef = self.ef

        if self.max_mem_queue_size:
            params.maxMemQueueSize = self.max_mem_queue_size

        params.batchingParams.CopyFrom(self.batching_params._to_pb2())
        params.cachingParams.CopyFrom(self.caching_params._to_pb2())

        params.healerParams.CopyFrom(self.healer_params._to_pb2())

        params.mergeParams.CopyFrom(self.merge_params._to_pb2())

        return params

    def __repr__(self) -> str:
        batching_repr = repr(self.batching_params)
        return (
            f"HnswParams(m={self.m}, ef_construction={self.ef_construction}, "
            f"ef={self.ef}, {batching_repr})"
        )

    def __str__(self) -> str:
        batching_str = str(self.batching_params)
        return (
            f"hnswParams {{\n"
            f"  m: {self.m}\n"
            f"  efConstruction: {self.ef_construction}\n"
            f"  ef: {self.ef}\n"
            f"  {batching_str}\n"
            f"}}"
        )

    def __getitem__(self, key):
        key = str(key)  # Ensure key is a string
        if not hasattr(self, key):
            raise AttributeError(f"'IndexDefinition' object has no attribute '{key}'")
        return getattr(self, key)

    def __setitem__(self, key, value):
        return setattr(self, key, value)


class HnswSearchParams(object):
    """
    Parameters for Hierarchical Navigable Small World (HNSW) search.

    HNSW is an algorithm used for approximate nearest neighbor search.

    :param ef: The parameter 'ef' controls the trade-off between search quality and search efficiency.
        It determines the size of the dynamic list of nearest neighbors (candidates) examined during the
        search phase. Larger values of 'ef' typically yield higher recall but slower search times.
        Defaults to None, meaning the algorithm uses a library-defined default value.
    :type ef: Optional[int]

    Notes:
        - 'ef' stands for "exploration factor."
        - Setting 'ef' to a higher value increases the recall (i.e., the likelihood of finding the true nearest neighbors) at the cost of increased computational overhead during the search process.

    """

    def __init__(self, *, ef: Optional[int] = None) -> None:

        self.ef = ef

    def _to_pb2(self):
        params = types_pb2.HnswSearchParams()
        params.ef = self.ef
        return params


class IndexStorage(object):
    """
    Helper class primarily used to specify which namespace and set to build the index on.

    :param namespace: The name of the namespace to build the index on.
    :type namespace: str

    :param set_name: The name of the set to build the index on. Defaults to None.
    :type set_name: Optional[str]

    """

    def __init__(
        self, *, namespace: Optional[str], set_name: Optional[str] = None
    ) -> None:
        self.namespace = namespace
        self.set_name = set_name

    def _to_pb2(self):
        index_storage = types_pb2.IndexStorage()
        if self.namespace:

            index_storage.namespace = self.namespace
        if self.set_name:
            index_storage.set = self.set_name
        return index_storage

    def __repr__(self) -> str:
        return (
            f"IndexStorage(namespace={self.namespace!r}, "
            f"set_name={self.set_name!r})"
        )

    def __str__(self) -> str:
        return (
            f"IndexStorage {{\n"
            f"  namespace: {self.namespace}\n"
            f"  set_name: {self.set_name}\n"
            f"}}"
        )

    def __getitem__(self, key):
        key = str(key)  # Ensure key is a string
        if not hasattr(self, key):
            raise AttributeError(f"'IndexDefinition' object has no attribute '{key}'")
        return getattr(self, key)

    def __setitem__(self, key, value):
        return setattr(self, key, value)


class IndexId(object):
    """
    AVS IndexId used in :class:`IndexDefintion`

    :param namespace: The name of the namespace in which to index records.
    :type namespace: str

    :param set_name: The name of the index.
    :type set_name: Optional[str]

    """

    def __init__(self, *, namespace: str, name: str) -> None:
        self.namespace = namespace
        self.name = name

    def __repr__(self) -> str:
        return f"IndexId(namespace={self.namespace!r}, name={self.name!r})"

    def __str__(self) -> str:
        return f"IndexId(namespace={self.namespace}, name={self.name})"

    def __eq__(self, other) -> bool:
        if not isinstance(other, IndexId):
            return NotImplemented
        return self.namespace == other.namespace and self.name == other.name

    def __getitem__(self, key):
        key = str(key)  # Ensure key is a string
        if not hasattr(self, key):
            raise AttributeError(f"'IndexId' object has no attribute '{key}'")
        return getattr(self, key)

    def __setitem__(self, key, value):
        return setattr(self, key, value)


class IndexDefinition(object):
    """
    AVS Index Definition

    :param id: Index ID.
    :type id: str

    :param dimensions: Number of dimensions.
    :type dimensions: int

    :param vector_distance_metric: Metric used to evaluate vector searches on the given index
    :type vector_distance_metric: VectorDistanceMetric

    :param field: Field name.
    :type field: str

    :param sets: Set name
    :type sets: str
    
    :param hnsw_params: HNSW parameters.
    :type hnsw_params: HnswParams

    :param storage: Index storage details.
    :type storage: IndexStorage

    :param index_labels: Meta data associated with the index. Defaults to None.
    :type index_labels: Optional[dict[str, str]]
    """

    def __init__(
        self,
        *,
        id: str,
        dimensions: int,
        vector_distance_metric: types_pb2.VectorDistanceMetric,
        field: str,
        sets: str,
        hnsw_params: HnswParams,
        storage: IndexStorage,
        index_labels: dict[str, str]
    ) -> None:
        self.id = id
        self.dimensions = dimensions
        self.vector_distance_metric = vector_distance_metric
        self.field = field
        self.sets = sets
        self.hnsw_params = hnsw_params
        self.storage = storage
        self.index_labels = index_labels

    def __repr__(self) -> str:
        return (
            f"IndexDefinition(id={self.id!r}, dimensions={self.dimensions}, field={self.field!r}, sets={self.sets!r},"
            f"vector_distance_metric={self.vector_distance_metric!r}, hnsw_params={self.hnsw_params!r}, storage={self.storage!r}, "
            f"index_labels={self.index_labels}"
        )

    def __str__(self) -> str:
        return (
            f"IndexDefinition(id={self.id}, dimensions={self.dimensions}, field={self.field}, sets={self.sets!r}, "
            f"vector_distance_metric={self.vector_distance_metric}, hnsw_params={self.hnsw_params}, storage={self.storage}, "
            f"index_labels={self.index_labels}"
        )

    def __eq__(self, other) -> bool:
        if not isinstance(other, IndexDefinition):
            return NotImplemented
        return (
            self.id == other.id
            and self.dimensions == other.dimensions
            and self.vector_distance_metric == other.vector_distance_metric
            and self.field == other.field
            and self.sets == other.sets
            and self.hnsw_params == other.hnsw_params
            and self.storage == other.storage
            and self.index_labels == other.index_labels
        )

    def __getitem__(self, key):
        key = str(key)  # Ensure key is a string
        if not hasattr(self, key):
            raise AttributeError(f"'IndexDefinition' object has no attribute '{key}'")
        return getattr(self, key)

    def __setitem__(self, key, value):
        return setattr(self, key, value)


class AVSError(Exception):
    """
    Custom exception raised for errors related to AVS.
    """

    pass


class AVSServerError(AVSError):
    """
    Custom exception raised for errors related to the AVS server.

    :param rpc_error: exception thrown by the grpc Python library on server calls. Defaults to None.
    :type set_name: grpc.RpcError

    """

    def __init__(self, *, rpc_error) -> None:
        self.rpc_error = rpc_error

    def __str__(self):
        return f"AVSServerError(rpc_error={self.rpc_error})"


class AVSClientError(AVSError):
    """
    Custom exception raised for errors related to AVS client-side failures..

    :param message: error messaging raised by the AVS Client. Defaults to None.
    :type set_name: str

    """

    def __init__(self, *, message) -> None:
        self.message = message

    def __str__(self):
        return f"AVSClientError(message={self.message})"



import re
import asyncio
import logging
import jwt
from jwt.exceptions import InvalidTokenError
from typing import Optional, Union

import google.protobuf.empty_pb2
import grpc
import random

from ... import types
from ...shared.proto_generated import vector_db_pb2
from ...shared.proto_generated import vector_db_pb2_grpc
from ...shared import base_channel_provider

empty = google.protobuf.empty_pb2.Empty()

logger = logging.getLogger(__name__)

TEND_INTERVAL = 1

class ChannelProvider(base_channel_provider.BaseChannelProvider):
    """AVS Channel Provider"""

    def __init__(
        self,
        seeds: tuple[types.HostPort, ...],
        listener_name: Optional[str] = None,
        is_loadbalancer: Optional[bool] = False,
        username: Optional[str] = None,
        password: Optional[str] = None,
        root_certificate: Optional[Union[list[str], str]] = None,
        certificate_chain: Optional[str] = None,
        private_key: Optional[str] = None,
        service_config_path: Optional[str] = None,
    ) -> None:

        super().__init__(
            seeds,
            listener_name,
            is_loadbalancer,
            username,
            password,
            root_certificate,
            certificate_chain,
            private_key,
            service_config_path,
        )

        # When set, client has concluded cluster tending
        self._tend_ended: asyncio.Event = asyncio.Event()

        # When set, client has completed a cluster tend cycle, initialized auth, and verified client-server minimum compatibility
        self._ready: asyncio.Event = asyncio.Event()

        # When locked, new task is being assigned to _auth_task
        self._auth_tending_lock: asyncio.Lock = asyncio.Lock()

        # initializes authentication tending
        self._auth_task: Optional[asyncio.Task] = asyncio.create_task(self._tend_token())

        # initializes client tending processes
        asyncio.create_task(self._tend())

        # Exception to progotate to main control flow from errors generated during tending
        self._tend_exception: Exception = None

    async def _is_ready(self):
        # Wait 1 round of cluster tending, auth token initialization, and server client compatiblity verfication
        await self._ready.wait()

        # This propogates any fatal/unexpected errors from client initialization/tending to the client. 
        # Raising errors in a task does not deliver this error information to users
        if self._tend_exception:
            raise self._tend_exception

    async def _tend(self):

        try:
            await self._auth_task

            # verfies server is minimally compatible with client
            await self._check_server_version()

            await self._tend_cluster()

            self._ready.set()

        except Exception as e:
            # Set all event to prevent hanging if initial tend fails with error
            self._tend_ended.set()
            self._ready.set()

    async def _tend_cluster(self):
        try:
            (channels, end_tend_cluster) = self.init_tend_cluster()
            

            if end_tend_cluster:
                self._tend_ended.set()
                return

            (cluster_info_stubs, tasks) = self._gather_new_cluster_ids_and_cluster_info_stubs(channels)

            new_cluster_ids = await asyncio.gather(*tasks)

            update_endpoints_stubs = self._gather_stubs_for_endpoint_updating(new_cluster_ids, cluster_info_stubs)

            tasks = self._gather_temp_endpoints(new_cluster_ids, update_endpoints_stubs)

            cluster_endpoints_list = await asyncio.gather(*tasks)

            temp_endpoints = self._assign_temporary_endpoints(cluster_endpoints_list)

            if update_endpoints_stubs:
                
                tasks = self._add_new_channels_from_temp_endpoints(temp_endpoints)
                
                await asyncio.gather(*tasks)

                tasks = self._close_old_channels_from_node_channels(temp_endpoints)

                await asyncio.gather(*tasks)

            await asyncio.sleep(TEND_INTERVAL)


            asyncio.create_task(self._tend_cluster())

        except Exception as e:
            logger.error("Unexpected tend failure: %s", e)
            self._tend_exception = e
            raise e

    async def _get_cluster_id_coroutine(self, stub):
        try:
            return await stub.GetClusterId(empty, credentials=self._token,)
        except Exception as e:
            logger.debug(
                "While tending, failed to get cluster id with error: " + str(e)
            )

    async def _get_cluster_endpoints_coroutine(self, stub):
        try:
            return (
                await stub.GetClusterEndpoints(
                    vector_db_pb2.ClusterNodeEndpointsRequest(
                        listenerName=self.listener_name
                    ),
                    credentials=self._token,
                )
            ).endpoints
        except Exception as e:
            logger.debug(
                "While tending, failed to get cluster endpoints with error: "
                + str(e)
            )

    async def _close_on_channel_coroutine(self, channel_endpoints):
        try:
            await channel_endpoints.channel.close()
        except Exception as e:
            logger.debug(
                "While tending, failed to close GRPC channel: "
                + str(e)
            )

    def _call_get_cluster_id(self, stub):
        return asyncio.create_task(self._get_cluster_id_coroutine(stub))  

    def _call_get_cluster_endpoints(self, stub):
        return asyncio.create_task(self._get_cluster_endpoints_coroutine(stub))

    def _call_close_on_channel(self, channel_endpoints):
        return asyncio.create_task(self._close_on_channel_coroutine(channel_endpoints))

    async def _tend_token(self):
        try:
            if not self._token:
                return
            elif self._token != True:
                await asyncio.sleep((self._ttl * self._ttl_threshold))

            await self._update_token_and_ttl()

            async with self._auth_tending_lock:
                self._auth_task = asyncio.create_task(self._tend_token())

        except Exception as e: 
            self._tend_exception = e
            logger.error("Failed to tend token with error: %s", e)            
            raise e

    async def _update_token_and_ttl(
        self,
    ) -> None:

        (auth_stub, auth_request) = self._prepare_authenticate(
            self._credentials, logger
        )

        try:
            response = await auth_stub.Authenticate(auth_request)
        except grpc.RpcError as e:
            logger.error("Failed to refresh authentication token with error: %s", e)
            raise types.AVSServerError(rpc_error=e)

        self._respond_authenticate(response.token)

    async def _check_server_version(self):
        try: 
            stub = vector_db_pb2_grpc.AboutServiceStub(self.get_channel())
            about_request = vector_db_pb2.AboutRequest()

            try:
                response = await stub.Get(
                    about_request, credentials=self._token
                )
                self.current_server_version = response.version
            except grpc.RpcError as e:
                logger.debug(
                    "Failed to retrieve server version: "
                    + str(e)
                )
                self._tend_exception = AVSServerError(rpc_error=e)  
            self.verify_compatibile_server()

        except Exception as e:
            logger.debug(
                "Failed to retrieve server version: "
                + str(e)
            )
            self._tend_exception = e
            raise e

    def _create_channel(self, host: str, port: int) -> grpc.Channel:
        host = re.sub(r"%.*", "", host)

        if self.service_config_json:
            options = []
            options.append(("grpc.service_config", self.service_config_json))
        else:
            options = None

        if self._root_certificate:

            ssl_credentials = grpc.ssl_channel_credentials(
                root_certificates=self._root_certificate,
                certificate_chain=self._certificate_chain,
                private_key=self._private_key,
            )

            return grpc.aio.secure_channel(
                f"{host}:{port}", ssl_credentials, options=options
            )

        else:
            return grpc.aio.insecure_channel(f"{host}:{port}", options=options)

    async def close(self):
        # signals to tend_cluster to end cluster tending
        self._closed = True

        # wait until cluster tending has ended
        await self._tend_ended.wait()

        for channel in self._seedChannels:
            await channel.close()

        for k, channelEndpoints in self._node_channels.items():
            if channelEndpoints.channel:
                await channelEndpoints.channel.close()

        async with self._auth_tending_lock:
            self._auth_task.cancel()


# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from . import vector_db_pb2 as vector__db__pb2

GRPC_GENERATED_VERSION = '1.64.1'
GRPC_VERSION = grpc.__version__
EXPECTED_ERROR_RELEASE = '1.65.0'
SCHEDULED_RELEASE_DATE = 'June 25, 2024'
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    warnings.warn(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in vector_db_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
        + f' This warning will become an error in {EXPECTED_ERROR_RELEASE},'
        + f' scheduled for release on {SCHEDULED_RELEASE_DATE}.',
        RuntimeWarning
    )


class AboutServiceStub(object):
    """Information about the service.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Get = channel.unary_unary(
                '/aerospike.vector.AboutService/Get',
                request_serializer=vector__db__pb2.AboutRequest.SerializeToString,
                response_deserializer=vector__db__pb2.AboutResponse.FromString,
                _registered_method=True)


class AboutServiceServicer(object):
    """Information about the service.
    """

    def Get(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_AboutServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Get': grpc.unary_unary_rpc_method_handler(
                    servicer.Get,
                    request_deserializer=vector__db__pb2.AboutRequest.FromString,
                    response_serializer=vector__db__pb2.AboutResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'aerospike.vector.AboutService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('aerospike.vector.AboutService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class AboutService(object):
    """Information about the service.
    """

    @staticmethod
    def Get(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/aerospike.vector.AboutService/Get',
            vector__db__pb2.AboutRequest.SerializeToString,
            vector__db__pb2.AboutResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)


class ClusterInfoServiceStub(object):
    """Vector DB cluster service.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetNodeId = channel.unary_unary(
                '/aerospike.vector.ClusterInfoService/GetNodeId',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=vector__db__pb2.NodeId.FromString,
                _registered_method=True)
        self.GetClusterId = channel.unary_unary(
                '/aerospike.vector.ClusterInfoService/GetClusterId',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=vector__db__pb2.ClusterId.FromString,
                _registered_method=True)
        self.GetClusteringState = channel.unary_unary(
                '/aerospike.vector.ClusterInfoService/GetClusteringState',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=vector__db__pb2.ClusteringState.FromString,
                _registered_method=True)
        self.GetClusterEndpoints = channel.unary_unary(
                '/aerospike.vector.ClusterInfoService/GetClusterEndpoints',
                request_serializer=vector__db__pb2.ClusterNodeEndpointsRequest.SerializeToString,
                response_deserializer=vector__db__pb2.ClusterNodeEndpoints.FromString,
                _registered_method=True)


class ClusterInfoServiceServicer(object):
    """Vector DB cluster service.
    """

    def GetNodeId(self, request, context):
        """Get the internal cluster node-Id for this server.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetClusterId(self, request, context):
        """Get current cluster-Id for the current cluster.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetClusteringState(self, request, context):
        """Get current cluster-Id for the current cluster.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetClusterEndpoints(self, request, context):
        """Get the advertised/listening endpoints for all nodes in the cluster, given a listener name.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ClusterInfoServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetNodeId': grpc.unary_unary_rpc_method_handler(
                    servicer.GetNodeId,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=vector__db__pb2.NodeId.SerializeToString,
            ),
            'GetClusterId': grpc.unary_unary_rpc_method_handler(
                    servicer.GetClusterId,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=vector__db__pb2.ClusterId.SerializeToString,
            ),
            'GetClusteringState': grpc.unary_unary_rpc_method_handler(
                    servicer.GetClusteringState,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=vector__db__pb2.ClusteringState.SerializeToString,
            ),
            'GetClusterEndpoints': grpc.unary_unary_rpc_method_handler(
                    servicer.GetClusterEndpoints,
                    request_deserializer=vector__db__pb2.ClusterNodeEndpointsRequest.FromString,
                    response_serializer=vector__db__pb2.ClusterNodeEndpoints.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'aerospike.vector.ClusterInfoService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('aerospike.vector.ClusterInfoService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class ClusterInfoService(object):
    """Vector DB cluster service.
    """

    @staticmethod
    def GetNodeId(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/aerospike.vector.ClusterInfoService/GetNodeId',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            vector__db__pb2.NodeId.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetClusterId(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/aerospike.vector.ClusterInfoService/GetClusterId',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            vector__db__pb2.ClusterId.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetClusteringState(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/aerospike.vector.ClusterInfoService/GetClusteringState',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            vector__db__pb2.ClusteringState.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetClusterEndpoints(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/aerospike.vector.ClusterInfoService/GetClusterEndpoints',
            vector__db__pb2.ClusterNodeEndpointsRequest.SerializeToString,
            vector__db__pb2.ClusterNodeEndpoints.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



from typing import Any, Optional, Union
import time
import numpy as np
from . import conversions

from .proto_generated import transact_pb2
from .proto_generated import transact_pb2_grpc
from .. import types
from .proto_generated import types_pb2
from . import helpers


class BaseClient(object):

    def _prepare_seeds(self, seeds) -> None:
        return helpers._prepare_seeds(seeds)

    def _prepare_put(
        self,
        namespace,
        key,
        record_data,
        set_name,
        write_type,
        ignore_mem_queue_full,
        timeout,
        logger,
    ) -> None:

        logger.debug(
            "Putting record: namespace=%s, key=%s, record_data:%s, set_name:%s, ignore_mem_queue_full %s, timeout:%s",
            namespace,
            key,
            record_data,
            set_name,
            ignore_mem_queue_full,
            timeout,
        )

        kwargs = {}
        if timeout is not None:
            kwargs["timeout"] = timeout

        key = self._get_key(namespace, set_name, key)
        field_list = []

        for k, v in record_data.items():
            if isinstance(v, np.ndarray):
                field_list.append(
                    types_pb2.Field(
                        name=k, value=conversions.toVectorDbValue(v.tolist())
                    )
                )

            else:
                field_list.append(
                    types_pb2.Field(name=k, value=conversions.toVectorDbValue(v))
                )

        transact_stub = self._get_transact_stub()
        put_request = transact_pb2.PutRequest(
            key=key,
            writeType=write_type,
            fields=field_list,
            ignoreMemQueueFull=ignore_mem_queue_full,
        )

        return (transact_stub, put_request, kwargs)

    def _prepare_insert(
        self,
        namespace,
        key,
        record_data,
        set_name,
        ignore_mem_queue_full,
        timeout,
        logger,
    ) -> None:
        return self._prepare_put(
            namespace,
            key,
            record_data,
            set_name,
            transact_pb2.WriteType.INSERT_ONLY,
            ignore_mem_queue_full,
            timeout,
            logger,
        )

    def _prepare_update(
        self,
        namespace,
        key,
        record_data,
        set_name,
        ignore_mem_queue_full,
        timeout,
        logger,
    ) -> None:
        return self._prepare_put(
            namespace,
            key,
            record_data,
            set_name,
            transact_pb2.WriteType.UPDATE_ONLY,
            ignore_mem_queue_full,
            timeout,
            logger,
        )

    def _prepare_upsert(
        self,
        namespace,
        key,
        record_data,
        set_name,
        ignore_mem_queue_full,
        timeout,
        logger,
    ) -> None:
        return self._prepare_put(
            namespace,
            key,
            record_data,
            set_name,
            transact_pb2.WriteType.UPSERT,
            ignore_mem_queue_full,
            timeout,
            logger,
        )

    def _prepare_get(
        self, namespace, key, field_names, set_name, timeout, logger
    ) -> None:

        logger.debug(
            "Getting record: namespace=%s, key=%s, field_names:%s, set_name:%s, timeout:%s",
            namespace,
            key,
            field_names,
            set_name,
            timeout,
        )

        kwargs = {}
        if timeout is not None:
            kwargs["timeout"] = timeout

        key = self._get_key(namespace, set_name, key)
        projection_spec = self._get_projection_spec(field_names=field_names)

        transact_stub = self._get_transact_stub()
        get_request = transact_pb2.GetRequest(key=key, projection=projection_spec)

        return (transact_stub, key, get_request, kwargs)

    def _prepare_exists(self, namespace, key, set_name, timeout, logger) -> None:

        logger.debug(
            "Getting record existence: namespace=%s, key=%s, set_name:%s, timeout:%s",
            namespace,
            key,
            set_name,
            timeout,
        )

        kwargs = {}
        if timeout is not None:
            kwargs["timeout"] = timeout

        key = self._get_key(namespace, set_name, key)

        transact_stub = self._get_transact_stub()
        exists_request = transact_pb2.ExistsRequest(key=key)

        return (transact_stub, exists_request, kwargs)

    def _prepare_delete(self, namespace, key, set_name, timeout, logger) -> None:

        logger.debug(
            "Deleting record: namespace=%s, key=%s, set_name=%s, timeout:%s",
            namespace,
            key,
            set_name,
            timeout,
        )

        kwargs = {}
        if timeout is not None:
            kwargs["timeout"] = timeout

        key = self._get_key(namespace, set_name, key)

        transact_stub = self._get_transact_stub()
        delete_request = transact_pb2.DeleteRequest(key=key)

        return (transact_stub, delete_request, kwargs)

    def _prepare_is_indexed(
        self, namespace, key, index_name, index_namespace, set_name, timeout, logger
    ) -> None:

        kwargs = {}
        if timeout is not None:
            kwargs["timeout"] = timeout

        logger.debug(
            "Checking if index exists: namespace=%s, key=%s, index_name=%s, index_namespace=%s, set_name=%s, timeout:%s",
            namespace,
            key,
            index_name,
            index_namespace,
            set_name,
            timeout,
        )

        if not index_namespace:
            index_namespace = namespace
        index_id = types_pb2.IndexId(namespace=index_namespace, name=index_name)
        key = self._get_key(namespace, set_name, key)

        transact_stub = self._get_transact_stub()
        is_indexed_request = transact_pb2.IsIndexedRequest(key=key, indexId=index_id)

        return (transact_stub, is_indexed_request, kwargs)

    def _prepare_vector_search(
        self,
        namespace,
        index_name,
        query,
        limit,
        search_params,
        field_names,
        timeout,
        logger,
    ) -> None:

        kwargs = {}
        if timeout is not None:
            kwargs["timeout"] = timeout

        logger.debug(
            "Performing vector search: namespace=%s, index_name=%s, query=%s, limit=%s, search_params=%s, field_names=%s, timeout:%s",
            namespace,
            index_name,
            query,
            limit,
            search_params,
            field_names,
            timeout,
        )

        if search_params != None:
            search_params = search_params._to_pb2()

        projection_spec = self._get_projection_spec(field_names=field_names)

        index = types_pb2.IndexId(namespace=namespace, name=index_name)

        if isinstance(query, np.ndarray):
            query_vector = conversions.toVectorDbValue(query.tolist()).vectorValue
        else:
            query_vector = conversions.toVectorDbValue(query).vectorValue

        transact_stub = self._get_transact_stub()

        vector_search_request = transact_pb2.VectorSearchRequest(
            index=index,
            queryVector=query_vector,
            limit=limit,
            hnswSearchParams=search_params,
            projection=projection_spec,
        )

        return (transact_stub, vector_search_request, kwargs)

    def _get_transact_stub(self):
        return transact_pb2_grpc.TransactServiceStub(
            self._channel_provider.get_channel()
        )

    def _respond_get(self, response, key) -> None:
        return types.RecordWithKey(
            key=conversions.fromVectorDbKey(key),
            fields=conversions.fromVectorDbRecord(response),
        )

    def _respond_exists(self, response) -> None:
        return response.value

    def _respond_is_indexed(self, response) -> None:
        return response.value

    def _respond_neighbor(self, response) -> None:
        return conversions.fromVectorDbNeighbor(response)

    def _get_projection_spec(
        self,
        *,
        field_names: Optional[list] = None,
        exclude_field_names: Optional[list] = None,
    ):

        if field_names:
            include = transact_pb2.ProjectionFilter(
                type=transact_pb2.ProjectionType.SPECIFIED, fields=field_names
            )
            exclude = transact_pb2.ProjectionFilter(
                type=transact_pb2.ProjectionType.NONE, fields=None
            )
        elif exclude_field_names:
            include = transact_pb2.ProjectionFilter(
                type=transact_pb2.ProjectionType.NONE, fields=None
            )
            exclude = transact_pb2.ProjectionFilter(
                type=transact_pb2.ProjectionType.SPECIFIED, fields=exclude_field_names
            )
        else:
            include = transact_pb2.ProjectionFilter(
                type=transact_pb2.ProjectionType.ALL, fields=None
            )
            exclude = transact_pb2.ProjectionFilter(
                type=transact_pb2.ProjectionType.NONE, fields=None
            )

        projection_spec = transact_pb2.ProjectionSpec(include=include, exclude=exclude)

        return projection_spec

    def _get_key(
        self, namespace: str, set: str, key: Union[int, str, bytes, bytearray]
    ):

        if isinstance(key, np.ndarray):
            key = key.tobytes()

        if isinstance(key, np.generic):
            key = key.item()

        if isinstance(key, str):
            key = types_pb2.Key(namespace=namespace, set=set, stringValue=key)
        elif isinstance(key, int):
            key = types_pb2.Key(namespace=namespace, set=set, longValue=key)
        elif isinstance(key, (bytes, bytearray)):
            key = types_pb2.Key(namespace=namespace, set=set, bytesValue=key)
        else:
            raise Exception("Invalid key type" + str(type(key)))
        return key

    def _prepare_wait_for_index_waiting(self, namespace, name, wait_interval):
        return helpers._prepare_wait_for_index_waiting(
            self, namespace, name, wait_interval
        )

    def _check_timeout(self, start_time, timeout):
        if start_time + timeout < time.monotonic():
            raise AVSClientError(message="timed-out waiting for index creation")
            
    def _check_completion_condition(
        self, start_time, timeout, index_status, unmerged_record_initialized
    ):
        self._check_timeout(start_time, timeout)

        if start_time + 10 < time.monotonic():
            unmerged_record_initialized = True

        if index_status.unmergedRecordCount > 0:
            unmerged_record_initialized = True

        if (
            index_status.unmergedRecordCount == 0
            and unmerged_record_initialized == True
        ):
            return True
        else:
            return False



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: types.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0btypes.proto\x12\x10\x61\x65rospike.vector\"\x91\x01\n\x03Key\x12\x11\n\tnamespace\x18\x01 \x01(\t\x12\x10\n\x03set\x18\x02 \x01(\tH\x01\x88\x01\x01\x12\x15\n\x0bstringValue\x18\x03 \x01(\tH\x00\x12\x14\n\nbytesValue\x18\x04 \x01(\x0cH\x00\x12\x12\n\x08intValue\x18\x05 \x01(\x05H\x00\x12\x13\n\tlongValue\x18\x06 \x01(\x03H\x00\x42\x07\n\x05valueB\x06\n\x04_set\"\x19\n\x08\x42oolData\x12\r\n\x05value\x18\x01 \x03(\x08\"\x1a\n\tFloatData\x12\r\n\x05value\x18\x01 \x03(\x02\"\x94\x01\n\x06MapKey\x12\x15\n\x0bstringValue\x18\x01 \x01(\tH\x00\x12\x14\n\nbytesValue\x18\x02 \x01(\x0cH\x00\x12\x12\n\x08intValue\x18\x03 \x01(\x05H\x00\x12\x13\n\tlongValue\x18\x04 \x01(\x03H\x00\x12\x14\n\nfloatValue\x18\x05 \x01(\x02H\x00\x12\x15\n\x0b\x64oubleValue\x18\x06 \x01(\x01H\x00\x42\x07\n\x05value\"Y\n\x08MapEntry\x12%\n\x03key\x18\x01 \x01(\x0b\x32\x18.aerospike.vector.MapKey\x12&\n\x05value\x18\x02 \x01(\x0b\x32\x17.aerospike.vector.Value\"2\n\x03Map\x12+\n\x07\x65ntries\x18\x01 \x03(\x0b\x32\x1a.aerospike.vector.MapEntry\"0\n\x04List\x12(\n\x07\x65ntries\x18\x01 \x03(\x0b\x32\x17.aerospike.vector.Value\"r\n\x06Vector\x12.\n\x08\x62oolData\x18\x01 \x01(\x0b\x32\x1a.aerospike.vector.BoolDataH\x00\x12\x30\n\tfloatData\x18\x02 \x01(\x0b\x32\x1b.aerospike.vector.FloatDataH\x00\x42\x06\n\x04\x64\x61ta\"\xb4\x02\n\x05Value\x12\x15\n\x0bstringValue\x18\x01 \x01(\tH\x00\x12\x14\n\nbytesValue\x18\x02 \x01(\x0cH\x00\x12\x12\n\x08intValue\x18\x03 \x01(\x05H\x00\x12\x13\n\tlongValue\x18\x04 \x01(\x03H\x00\x12\x14\n\nfloatValue\x18\x05 \x01(\x02H\x00\x12\x15\n\x0b\x64oubleValue\x18\x06 \x01(\x01H\x00\x12)\n\x08mapValue\x18\x07 \x01(\x0b\x32\x15.aerospike.vector.MapH\x00\x12+\n\tlistValue\x18\x08 \x01(\x0b\x32\x16.aerospike.vector.ListH\x00\x12/\n\x0bvectorValue\x18\t \x01(\x0b\x32\x18.aerospike.vector.VectorH\x00\x12\x16\n\x0c\x62ooleanValue\x18\n \x01(\x08H\x00\x42\x07\n\x05value\"=\n\x05\x46ield\x12\x0c\n\x04name\x18\x01 \x01(\t\x12&\n\x05value\x18\x02 \x01(\x0b\x32\x17.aerospike.vector.Value\"A\n\x17\x41\x65rospikeRecordMetadata\x12\x12\n\ngeneration\x18\x01 \x01(\r\x12\x12\n\nexpiration\x18\x02 \x01(\r\"\x85\x01\n\x06Record\x12\'\n\x06\x66ields\x18\x01 \x03(\x0b\x32\x17.aerospike.vector.Field\x12\x46\n\x11\x61\x65rospikeMetadata\x18\x02 \x01(\x0b\x32).aerospike.vector.AerospikeRecordMetadataH\x00\x42\n\n\x08metadata\"z\n\x08Neighbor\x12\"\n\x03key\x18\x01 \x01(\x0b\x32\x15.aerospike.vector.Key\x12-\n\x06record\x18\x02 \x01(\x0b\x32\x18.aerospike.vector.RecordH\x00\x88\x01\x01\x12\x10\n\x08\x64istance\x18\x03 \x01(\x02\x42\t\n\x07_record\"*\n\x07IndexId\x12\x11\n\tnamespace\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\"\x8d\x03\n\nHnswParams\x12\x0e\n\x01m\x18\x01 \x01(\rH\x00\x88\x01\x01\x12\x1b\n\x0e\x65\x66\x43onstruction\x18\x02 \x01(\rH\x01\x88\x01\x01\x12\x0f\n\x02\x65\x66\x18\x03 \x01(\rH\x02\x88\x01\x01\x12<\n\x0e\x62\x61tchingParams\x18\x04 \x01(\x0b\x32$.aerospike.vector.HnswBatchingParams\x12\x1c\n\x0fmaxMemQueueSize\x18\x05 \x01(\rH\x03\x88\x01\x01\x12:\n\rcachingParams\x18\x06 \x01(\x0b\x32#.aerospike.vector.HnswCachingParams\x12\x38\n\x0chealerParams\x18\x07 \x01(\x0b\x32\".aerospike.vector.HnswHealerParams\x12;\n\x0bmergeParams\x18\x08 \x01(\x0b\x32&.aerospike.vector.HnswIndexMergeParamsB\x04\n\x02_mB\x11\n\x0f_efConstructionB\x05\n\x03_efB\x12\n\x10_maxMemQueueSize\"*\n\x10HnswSearchParams\x12\x0f\n\x02\x65\x66\x18\x01 \x01(\rH\x00\x88\x01\x01\x42\x05\n\x03_ef\"@\n\x14HnswIndexMergeParams\x12\x18\n\x0bparallelism\x18\x01 \x01(\rH\x00\x88\x01\x01\x42\x0e\n\x0c_parallelism\"[\n\x11HnswCachingParams\x12\x17\n\nmaxEntries\x18\x01 \x01(\x04H\x00\x88\x01\x01\x12\x13\n\x06\x65xpiry\x18\x02 \x01(\x04H\x01\x88\x01\x01\x42\r\n\x0b_maxEntriesB\t\n\x07_expiry\"\x84\x02\n\x10HnswHealerParams\x12\x1f\n\x12maxScanRatePerNode\x18\x01 \x01(\rH\x00\x88\x01\x01\x12\x1c\n\x0fmaxScanPageSize\x18\x02 \x01(\rH\x01\x88\x01\x01\x12\x1b\n\x0ereindexPercent\x18\x03 \x01(\x02H\x02\x88\x01\x01\x12\x1a\n\rscheduleDelay\x18\x04 \x01(\x04H\x03\x88\x01\x01\x12\x18\n\x0bparallelism\x18\x05 \x01(\rH\x04\x88\x01\x01\x42\x15\n\x13_maxScanRatePerNodeB\x12\n\x10_maxScanPageSizeB\x11\n\x0f_reindexPercentB\x10\n\x0e_scheduleDelayB\x0e\n\x0c_parallelism\"`\n\x12HnswBatchingParams\x12\x17\n\nmaxRecords\x18\x01 \x01(\rH\x00\x88\x01\x01\x12\x15\n\x08interval\x18\x02 \x01(\rH\x01\x88\x01\x01\x42\r\n\x0b_maxRecordsB\x0b\n\t_interval\"\x8e\x03\n\x0fHnswIndexUpdate\x12\x41\n\x0e\x62\x61tchingParams\x18\x01 \x01(\x0b\x32$.aerospike.vector.HnswBatchingParamsH\x00\x88\x01\x01\x12\x1c\n\x0fmaxMemQueueSize\x18\x02 \x01(\rH\x01\x88\x01\x01\x12?\n\rcachingParams\x18\x03 \x01(\x0b\x32#.aerospike.vector.HnswCachingParamsH\x02\x88\x01\x01\x12=\n\x0chealerParams\x18\x04 \x01(\x0b\x32\".aerospike.vector.HnswHealerParamsH\x03\x88\x01\x01\x12@\n\x0bmergeParams\x18\x05 \x01(\x0b\x32&.aerospike.vector.HnswIndexMergeParamsH\x04\x88\x01\x01\x42\x11\n\x0f_batchingParamsB\x12\n\x10_maxMemQueueSizeB\x10\n\x0e_cachingParamsB\x0f\n\r_healerParamsB\x0e\n\x0c_mergeParams\"N\n\x0cIndexStorage\x12\x16\n\tnamespace\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x10\n\x03set\x18\x02 \x01(\tH\x01\x88\x01\x01\x42\x0c\n\n_namespaceB\x06\n\x04_set\"\xe0\x03\n\x0fIndexDefinition\x12%\n\x02id\x18\x01 \x01(\x0b\x32\x19.aerospike.vector.IndexId\x12)\n\x04type\x18\x02 \x01(\x0e\x32\x1b.aerospike.vector.IndexType\x12\x12\n\ndimensions\x18\x03 \x01(\r\x12\x44\n\x14vectorDistanceMetric\x18\x04 \x01(\x0e\x32&.aerospike.vector.VectorDistanceMetric\x12\r\n\x05\x66ield\x18\x05 \x01(\t\x12\x16\n\tsetFilter\x18\x06 \x01(\tH\x01\x88\x01\x01\x12\x32\n\nhnswParams\x18\x07 \x01(\x0b\x32\x1c.aerospike.vector.HnswParamsH\x00\x12=\n\x06labels\x18\x08 \x03(\x0b\x32-.aerospike.vector.IndexDefinition.LabelsEntry\x12\x34\n\x07storage\x18\t \x01(\x0b\x32\x1e.aerospike.vector.IndexStorageH\x02\x88\x01\x01\x1a-\n\x0bLabelsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\x08\n\x06paramsB\x0c\n\n_setFilterB\n\n\x08_storage\"I\n\x13IndexDefinitionList\x12\x32\n\x07indices\x18\x01 \x03(\x0b\x32!.aerospike.vector.IndexDefinition\"\x12\n\x04Role\x12\n\n\x02id\x18\x01 \x01(\t\"\'\n\x04User\x12\x10\n\x08username\x18\x01 \x01(\t\x12\r\n\x05roles\x18\x02 \x03(\t\"t\n\x0b\x43redentials\x12\x10\n\x08username\x18\x01 \x01(\t\x12\x44\n\x13passwordCredentials\x18\x02 \x01(\x0b\x32%.aerospike.vector.PasswordCredentialsH\x00\x42\r\n\x0b\x63redentials\"\'\n\x13PasswordCredentials\x12\x10\n\x08password\x18\x01 \x01(\t\"\x18\n\x07\x42oolean\x12\r\n\x05value\x18\x01 \x01(\x08*f\n\x14VectorDistanceMetric\x12\x15\n\x11SQUARED_EUCLIDEAN\x10\x00\x12\n\n\x06\x43OSINE\x10\x01\x12\x0f\n\x0b\x44OT_PRODUCT\x10\x02\x12\r\n\tMANHATTAN\x10\x03\x12\x0b\n\x07HAMMING\x10\x04*\x15\n\tIndexType\x12\x08\n\x04HNSW\x10\x00\x42\x43\n!com.aerospike.vector.client.protoP\x01Z\x1c\x61\x65rospike.com/vector/protos/b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'types_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n!com.aerospike.vector.client.protoP\001Z\034aerospike.com/vector/protos/'
  _globals['_INDEXDEFINITION_LABELSENTRY']._loaded_options = None
  _globals['_INDEXDEFINITION_LABELSENTRY']._serialized_options = b'8\001'
  _globals['_VECTORDISTANCEMETRIC']._serialized_start=3690
  _globals['_VECTORDISTANCEMETRIC']._serialized_end=3792
  _globals['_INDEXTYPE']._serialized_start=3794
  _globals['_INDEXTYPE']._serialized_end=3815
  _globals['_KEY']._serialized_start=34
  _globals['_KEY']._serialized_end=179
  _globals['_BOOLDATA']._serialized_start=181
  _globals['_BOOLDATA']._serialized_end=206
  _globals['_FLOATDATA']._serialized_start=208
  _globals['_FLOATDATA']._serialized_end=234
  _globals['_MAPKEY']._serialized_start=237
  _globals['_MAPKEY']._serialized_end=385
  _globals['_MAPENTRY']._serialized_start=387
  _globals['_MAPENTRY']._serialized_end=476
  _globals['_MAP']._serialized_start=478
  _globals['_MAP']._serialized_end=528
  _globals['_LIST']._serialized_start=530
  _globals['_LIST']._serialized_end=578
  _globals['_VECTOR']._serialized_start=580
  _globals['_VECTOR']._serialized_end=694
  _globals['_VALUE']._serialized_start=697
  _globals['_VALUE']._serialized_end=1005
  _globals['_FIELD']._serialized_start=1007
  _globals['_FIELD']._serialized_end=1068
  _globals['_AEROSPIKERECORDMETADATA']._serialized_start=1070
  _globals['_AEROSPIKERECORDMETADATA']._serialized_end=1135
  _globals['_RECORD']._serialized_start=1138
  _globals['_RECORD']._serialized_end=1271
  _globals['_NEIGHBOR']._serialized_start=1273
  _globals['_NEIGHBOR']._serialized_end=1395
  _globals['_INDEXID']._serialized_start=1397
  _globals['_INDEXID']._serialized_end=1439
  _globals['_HNSWPARAMS']._serialized_start=1442
  _globals['_HNSWPARAMS']._serialized_end=1839
  _globals['_HNSWSEARCHPARAMS']._serialized_start=1841
  _globals['_HNSWSEARCHPARAMS']._serialized_end=1883
  _globals['_HNSWINDEXMERGEPARAMS']._serialized_start=1885
  _globals['_HNSWINDEXMERGEPARAMS']._serialized_end=1949
  _globals['_HNSWCACHINGPARAMS']._serialized_start=1951
  _globals['_HNSWCACHINGPARAMS']._serialized_end=2042
  _globals['_HNSWHEALERPARAMS']._serialized_start=2045
  _globals['_HNSWHEALERPARAMS']._serialized_end=2305
  _globals['_HNSWBATCHINGPARAMS']._serialized_start=2307
  _globals['_HNSWBATCHINGPARAMS']._serialized_end=2403
  _globals['_HNSWINDEXUPDATE']._serialized_start=2406
  _globals['_HNSWINDEXUPDATE']._serialized_end=2804
  _globals['_INDEXSTORAGE']._serialized_start=2806
  _globals['_INDEXSTORAGE']._serialized_end=2884
  _globals['_INDEXDEFINITION']._serialized_start=2887
  _globals['_INDEXDEFINITION']._serialized_end=3367
  _globals['_INDEXDEFINITION_LABELSENTRY']._serialized_start=3286
  _globals['_INDEXDEFINITION_LABELSENTRY']._serialized_end=3331
  _globals['_INDEXDEFINITIONLIST']._serialized_start=3369
  _globals['_INDEXDEFINITIONLIST']._serialized_end=3442
  _globals['_ROLE']._serialized_start=3444
  _globals['_ROLE']._serialized_end=3462
  _globals['_USER']._serialized_start=3464
  _globals['_USER']._serialized_end=3503
  _globals['_CREDENTIALS']._serialized_start=3505
  _globals['_CREDENTIALS']._serialized_end=3621
  _globals['_PASSWORDCREDENTIALS']._serialized_start=3623
  _globals['_PASSWORDCREDENTIALS']._serialized_end=3662
  _globals['_BOOLEAN']._serialized_start=3664
  _globals['_BOOLEAN']._serialized_end=3688
# @@protoc_insertion_point(module_scope)



from .client import Client
from .admin import Client as AdminClient
from ..types import (
    HostPort,
    Key,
    RecordWithKey,
    Neighbor,
    VectorDistanceMetric,
    HnswBatchingParams,
    HnswParams,
    HnswSearchParams,
)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: auth.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import types_pb2 as types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\nauth.proto\x12\x10\x61\x65rospike.vector\x1a\x0btypes.proto\"A\n\x0b\x41uthRequest\x12\x32\n\x0b\x63redentials\x18\x01 \x01(\x0b\x32\x1d.aerospike.vector.Credentials\"\x1d\n\x0c\x41uthResponse\x12\r\n\x05token\x18\x01 \x01(\t2^\n\x0b\x41uthService\x12O\n\x0c\x41uthenticate\x12\x1d.aerospike.vector.AuthRequest\x1a\x1e.aerospike.vector.AuthResponse\"\x00\x42\x43\n!com.aerospike.vector.client.protoP\x01Z\x1c\x61\x65rospike.com/vector/protos/b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'auth_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n!com.aerospike.vector.client.protoP\001Z\034aerospike.com/vector/protos/'
  _globals['_AUTHREQUEST']._serialized_start=45
  _globals['_AUTHREQUEST']._serialized_end=110
  _globals['_AUTHRESPONSE']._serialized_start=112
  _globals['_AUTHRESPONSE']._serialized_end=141
  _globals['_AUTHSERVICE']._serialized_start=143
  _globals['_AUTHSERVICE']._serialized_end=237
# @@protoc_insertion_point(module_scope)



import logging
import sys
import time
from typing import Any, Optional, Union

import grpc

from . import types
from .internal import channel_provider
from .shared.client_helpers import BaseClient

logger = logging.getLogger(__name__)


class Client(BaseClient):
    """
    Aerospike Vector Search Vector Client

    This client specializes in performing database operations with vector data.
    Moreover, the client supports Hierarchical Navigable Small World (HNSW) vector searches,
    allowing users to find vectors similar to a given query vector within an index.

    :param seeds: Defines the AVS nodes to which you want AVS to connect. AVS iterates through the seed nodes. After connecting to a node, AVS discovers all of the nodes in the cluster.
    :type seeds: Union[types.HostPort, tuple[types.HostPort, ...]]

    :param listener_name: An external (NATed) address and port combination that differs from the actual address and port where AVS is listening. Clients can access AVS on a node using the advertised listener address and port. Defaults to None.
    :type listener_name: Optional[str]

    :param is_loadbalancer: If true, the first seed address will be treated as a load balancer node. Defaults to False.
    :type is_loadbalancer: Optional[bool]

    :param service_config_path: Path to the service configuration file. Defaults to None.
    :type service_config_path: Optional[str]

    :param username: Username for Role-Based Access. Defaults to None.
    :type username: Optional[str]

    :param password: Password for Role-Based Access. Defaults to None.
    :type password: Optional[str]

    :param root_certificate: The PEM-encoded root certificates as a byte string. Defaults to None.
    :type root_certificate: Optional[list[bytes], bytes]

    :param certificate_chain: The PEM-encoded certificate chain as a byte string. Defaults to None.
    :type certificate_chain: Optional[bytes]

    :param private_key: The PEM-encoded private key as a byte string. Defaults to None.
    :type private_key: Optional[bytes]

    :raises AVSClientError: Raised when no seed host is provided.

    """

    def __init__(
        self,
        *,
        seeds: Union[types.HostPort, tuple[types.HostPort, ...]],
        listener_name: Optional[str] = None,
        is_loadbalancer: Optional[bool] = False,
        username: Optional[str] = None,
        password: Optional[str] = None,
        root_certificate: Optional[Union[list[str], str]] = None,
        certificate_chain: Optional[str] = None,
        private_key: Optional[str] = None,
        service_config_path: Optional[str] = None,
    ) -> None:

        seeds = self._prepare_seeds(seeds)
        self._channel_provider = channel_provider.ChannelProvider(
            seeds,
            listener_name,
            is_loadbalancer,
            username,
            password,
            root_certificate,
            certificate_chain,
            private_key,
            service_config_path,
        )

    def insert(
        self,
        *,
        namespace: str,
        key: Union[int, str, bytes, bytearray],
        record_data: dict[str, Any],
        set_name: Optional[str] = None,
        ignore_mem_queue_full: Optional[bool] = False,
        timeout: Optional[int] = None,
    ) -> None:
        """
        Insert a record into Aerospike Vector Search.

        If record does exist, an exception is raised.
        If record doesn't exist, the record is inserted.

        :param namespace: The namespace for the record.
        :type namespace: str

        :param key: The key for the record.
        :type key: Union[int, str, bytes, bytearray, np.generic, np.ndarray]

        :param record_data: The data to be stored in the record.
        :type record_data: dict[str, Any]

        :param set_name: The name of the set to which the record belongs. Defaults to None.
        :type set_name: Optional[str]

        :param ignore_mem_queue_full: Ignore the in-memory queue full error. These records will be written to storage
            and later, the index healer will pick them for indexing. Defaults to False.
        :type ignore_mem_queue_full: int

        :param timeout: Time in seconds this operation will wait before raising an :class:`AVSServerError <aerospike_vector_search.types.AVSServerError>`. Defaults to None.
        :type timeout: int

        Raises:
            AVSServerError: Raised if an error occurs during the RPC communication with the server while attempting to insert a vector.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.

        """

        (transact_stub, insert_request, kwargs) = self._prepare_insert(
            namespace,
            key,
            record_data,
            set_name,
            ignore_mem_queue_full,
            timeout,
            logger,
        )

        try:
            transact_stub.Put(
                insert_request, credentials=self._channel_provider.get_token(), **kwargs
            )
        except grpc.RpcError as e:
            logger.error("Failed to insert vector with error: %s", e)
            raise types.AVSServerError(rpc_error=e)

    def update(
        self,
        *,
        namespace: str,
        key: Union[int, str, bytes, bytearray],
        record_data: dict[str, Any],
        set_name: Optional[str] = None,
        ignore_mem_queue_full: Optional[bool] = False,
        timeout: Optional[int] = None,
    ) -> None:
        """
        Update a record in Aerospike Vector Search.

        :param namespace: The namespace for the record.
        :type namespace: str

        :param key: The key for the record.
        :type key: Union[int, str, bytes, bytearray, np.generic, np.ndarray]

        :param record_data: The data to be stored in the record.
        :type record_data: dict[str, Any]

        :param set_name: The name of the set to which the record belongs. Defaults to None.
        :type set_name: Optional[str]

        :param ignore_mem_queue_full: Ignore the in-memory queue full error. These records will be written to storage
            and later, the index healer will pick them for indexing. Defaults to False.
        :type ignore_mem_queue_full: int

        :param timeout: Time in seconds this operation will wait before raising an :class:`AVSServerError <aerospike_vector_search.types.AVSServerError>`. Defaults to None.
        :type timeout: int

        Raises:
            AVSServerError: Raised if an error occurs during the RPC communication with the server while attempting to update a vector.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.

        """
        (transact_stub, update_request, kwargs) = self._prepare_update(
            namespace,
            key,
            record_data,
            set_name,
            ignore_mem_queue_full,
            timeout,
            logger,
        )

        try:
            transact_stub.Put(
                update_request, credentials=self._channel_provider.get_token(), **kwargs
            )
        except grpc.RpcError as e:
            logger.error("Failed to update vector with error: %s", e)
            raise types.AVSServerError(rpc_error=e)

    def upsert(
        self,
        *,
        namespace: str,
        key: Union[int, str, bytes, bytearray],
        record_data: dict[str, Any],
        set_name: Optional[str] = None,
        ignore_mem_queue_full: Optional[bool] = False,
        timeout: Optional[int] = None,
    ) -> None:
        """
        Update a record in Aerospike Vector Search.

        If record does exist, update the record.
        If record doesn't exist, the record is inserted.

        :param namespace: The namespace for the record.
        :type namespace: str

        :param key: The key for the record.
        :type key: Union[int, str, bytes, bytearray, np.generic, np.ndarray]

        :param record_data: The data to be stored in the record.
        :type record_data: dict[str, Any]

        :param set_name: The name of the set to which the record belongs. Defaults to None.
        :type set_name: Optional[str]

        :param ignore_mem_queue_full: Ignore the in-memory queue full error. These records will be written to storage
            and later, the index healer will pick them for indexing. Defaults to False.
        :type ignore_mem_queue_full: int

        :param timeout: Time in seconds this operation will wait before raising an :class:`AVSServerError <aerospike_vector_search.types.AVSServerError>`. Defaults to None.
        :type timeout: int

        Raises:
            AVSServerError: Raised if an error occurs during the RPC communication with the server while attempting to upsert a vector.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.

        """

        (transact_stub, upsert_request, kwargs) = self._prepare_upsert(
            namespace,
            key,
            record_data,
            set_name,
            ignore_mem_queue_full,
            timeout,
            logger,
        )

        try:
            transact_stub.Put(
                upsert_request, credentials=self._channel_provider.get_token(), **kwargs
            )
        except grpc.RpcError as e:
            logger.error("Failed to upsert vector with error: %s", e)
            raise types.AVSServerError(rpc_error=e)

    def get(
        self,
        *,
        namespace: str,
        key: Union[int, str, bytes, bytearray],
        field_names: Optional[list[str]] = None,
        set_name: Optional[str] = None,
        timeout: Optional[int] = None,
    ) -> types.RecordWithKey:
        """
        Read a record from Aerospike Vector Search.

        :param namespace: The namespace for the record.
        :type namespace: str

        :param key: The key for the record.
        :type key: Union[int, str, bytes, bytearray, np.generic, np.ndarray]


        :param field_names: A list of field names to retrieve from the record.
            If None, all fields are retrieved. Defaults to None.
        :type field_names: Optional[list[str]]

        :param set_name: The name of the set from which to read the record. Defaults to None.
        :type set_name: Optional[str]

        :param timeout: Time in seconds this operation will wait before raising an :class:`AVSServerError <aerospike_vector_search.types.AVSServerError>`. Defaults to None.
        :type timeout: int

        Returns:
            types.RecordWithKey: A record with its associated key.

        Raises:
            AVSServerError: Raised if an error occurs during the RPC communication with the server while attempting to get a vector.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.
        """

        (transact_stub, key, get_request, kwargs) = self._prepare_get(
            namespace, key, field_names, set_name, timeout, logger
        )

        try:
            response = transact_stub.Get(
                get_request, credentials=self._channel_provider.get_token(), **kwargs
            )
        except grpc.RpcError as e:
            logger.error("Failed to get vector with error: %s", e)
            raise types.AVSServerError(rpc_error=e)

        return self._respond_get(response, key)

    def exists(
        self,
        *,
        namespace: str,
        key: Any,
        set_name: Optional[str] = None,
        timeout: Optional[int] = None,
    ) -> bool:
        """
        Check if a record exists in Aerospike Vector Search.

        :param namespace: The namespace for the record.
        :type namespace: str

        :param key: The key for the record.
        :type key: Union[int, str, bytes, bytearray, np.generic, np.ndarray]

        :param set_name: The name of the set to which the record belongs. Defaults to None.
        :type set_name: str

        :param timeout: Time in seconds this operation will wait before raising an :class:`AVSServerError <aerospike_vector_search.types.AVSServerError>`. Defaults to None.
        :type timeout: int

        Returns:
            bool: True if the record exists, False otherwise.

        Raises:
            AVSServerError: Raised if an error occurs during the RPC communication with the server while attempting to see if a given vector exists.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.
        """

        (transact_stub, exists_request, kwargs) = self._prepare_exists(
            namespace, key, set_name, timeout, logger
        )

        try:
            response = transact_stub.Exists(
                exists_request, credentials=self._channel_provider.get_token(), **kwargs
            )
        except grpc.RpcError as e:
            logger.error("Failed to verfiy vector existence with error: %s", e)
            raise types.AVSServerError(rpc_error=e)

        return self._respond_exists(response)

    def delete(
        self,
        *,
        namespace: str,
        key: Any,
        set_name: Optional[str] = None,
        timeout: Optional[int] = None,
    ) -> None:
        """
        Delete a record from Aerospike Vector Search.

        :param namespace: The namespace for the record.
        :type namespace: str

        :param key: The key for the record.
        :type key: Union[int, str, bytes, bytearray, np.generic, np.ndarray]

        :param set_name: The name of the set to which the record belongs. Defaults to None.
        :type set_name: Optional[str]

        :param timeout: Time in seconds this operation will wait before raising an :class:`AVSServerError <aerospike_vector_search.types.AVSServerError>`. Defaults to None.
        :type timeout: int

        Raises:
            AVSServerError: Raised if an error occurs during the RPC communication with the server while attempting to delete the index.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.
        """

        (transact_stub, delete_request, kwargs) = self._prepare_delete(
            namespace, key, set_name, timeout, logger
        )

        try:
            transact_stub.Delete(
                delete_request, credentials=self._channel_provider.get_token(), **kwargs
            )
        except grpc.RpcError as e:
            logger.error("Failed to delete vector with error: %s", e)
            raise types.AVSServerError(rpc_error=e)

    def is_indexed(
        self,
        *,
        namespace: str,
        key: Union[int, str, bytes, bytearray],
        index_name: str,
        index_namespace: Optional[str] = None,
        set_name: Optional[str] = None,
        timeout: Optional[int] = None,
    ) -> bool:
        """
        Check if a record is indexed in the Vector DB.

        :param namespace: The namespace for the record.
        :type namespace: str

        :param key: The key for the record.
        :type key: Union[int, str, bytes, bytearray, np.generic, np.ndarray]

        :param index_name: The name of the index.
        :type index_name: str

        :param index_namespace: The namespace of the index. If None, defaults to the namespace of the record. Defaults to None.
        :type index_namespace: optional[str]

        :param set_name: The name of the set to which the record belongs. Defaults to None.
        :type set_name: optional[str]

        :param timeout: Time in seconds this operation will wait before raising an :class:`AVSServerError <aerospike_vector_search.types.AVSServerError>`. Defaults to None.
        :type timeout: int

        Returns:
            bool: True if the record is indexed, False otherwise.

        Raises:
            AVSServerError: Raised if an error occurs during the RPC communication with the server while attempting to check if the vector is indexed.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.
        """

        (transact_stub, is_indexed_request, kwargs) = self._prepare_is_indexed(
            namespace, key, index_name, index_namespace, set_name, timeout, logger
        )

        try:
            response = transact_stub.IsIndexed(
                is_indexed_request,
                credentials=self._channel_provider.get_token(),
                **kwargs,
            )
        except grpc.RpcError as e:
            logger.error("Failed to verify vector indexing status with error: %s", e)
            raise types.AVSServerError(rpc_error=e)
        return self._respond_is_indexed(response)

    def vector_search(
        self,
        *,
        namespace: str,
        index_name: str,
        query: list[Union[bool, float]],
        limit: int,
        search_params: Optional[types.HnswSearchParams] = None,
        field_names: Optional[list[str]] = None,
        timeout: Optional[int] = None,
    ) -> list[types.Neighbor]:
        """
        Perform a Hierarchical Navigable Small World (HNSW) vector search in Aerospike Vector Search.

        :param namespace: The namespace for the records.
        :type namespace: str

        :param index_name: The name of the index.
        :type index_name: str

        :param query: The query vector for the search.
        :type query: list[Union[bool, float]]

        :param limit: The maximum number of neighbors to return. K value.
        :type limit: int

        :param search_params: Parameters for the HNSW algorithm.
            If None, the default parameters for the index are used. Defaults to None.
        :type search_params: Optional[types_pb2.HnswSearchParams]

        :param field_names: A list of field names to retrieve from the results.
            If None, all fields are retrieved. Defaults to None.
        :type field_names: Optional[list[str]]

        :param timeout: Time in seconds this operation will wait before raising an :class:`AVSServerError <aerospike_vector_search.types.AVSServerError>`. Defaults to None.
        :type timeout: int

        Returns:
            list[types.Neighbor]: A list of neighbors records found by the search.

        Raises:
            AVSServerError: Raised if an error occurs during the RPC communication with the server while attempting to vector search.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.
        """

        (transact_stub, vector_search_request, kwargs) = self._prepare_vector_search(
            namespace,
            index_name,
            query,
            limit,
            search_params,
            field_names,
            timeout,
            logger,
        )

        try:
            return [
                self._respond_neighbor(result)
                for result in transact_stub.VectorSearch(
                    vector_search_request,
                    credentials=self._channel_provider.get_token(),
                    **kwargs,
                )
            ]
        except grpc.RpcError as e:
            logger.error("Failed to vector search with error: %s", e)
            raise types.AVSServerError(rpc_error=e)

    def wait_for_index_completion(
        self,
        *,
        namespace: str,
        name: str,
        timeout: Optional[int] = sys.maxsize,
        wait_interval: Optional[int] = 12,
        validation_threshold: Optional[int] = 2,
    ) -> None:
        """
        Wait for the index to have no pending index update operations.

        :param namespace (str): The namespace of the index.
        :type namespace: str

        :param name (str): The name of the index.
        :type name: str

        :param timeout (int, optional): The maximum time (in seconds) to wait for the index to complete.
            Defaults to sys.maxsize.
        :type timeout: int

        :param wait_interval: The time (in seconds) to wait between index completion status request to the server.
            Lowering this value increases the chance that the index completion status is incorrect, which can result in poor search accuracy.
        :type wait_interval: int

        Raises:
            Exception: Raised when the timeout occurs while waiting for index completion.
            AVSServerError: Raised if an error occurs during the RPC communication with the server while attempting to wait for index completion.
            This error could occur due to various reasons such as network issues, server-side failures, or invalid request parameters.

        Note:
            The function polls the index status with a wait interval of 10 seconds until either
            the timeout is reached or the index has no pending index update operations.
        """
        (
            index_stub,
            wait_interval,
            start_time,
            unmerged_record_initialized,
            validation_count,
            index_completion_request,
        ) = self._prepare_wait_for_index_waiting(namespace, name, wait_interval)

        while True:
            try:
                index_status = index_stub.GetStatus(
                    index_completion_request,
                    credentials=self._channel_provider.get_token(),
                )
            except grpc.RpcError as e:

                logger.error("Failed waiting for index completion with error: %s", e)
                raise types.AVSServerError(rpc_error=e)
            if self._check_completion_condition(
                start_time, timeout, index_status, unmerged_record_initialized
            ):
                if validation_count == validation_threshold:
                    return
                else:
                    validation_count += 1
            else:
                validation_count = 0
            time.sleep(wait_interval)

    def close(self):
        """
        Close the Aerospike Vector Search Vector Client.

        This method closes gRPC channels connected to Aerospike Vector Search.

        Note:
            This method should be called when the VectorDbAdminClient is no longer needed to release resources.
        """
        self._channel_provider.close()

    def __enter__(self):
        """
        Enter a context manager for the vector client.

        Returns:
            VectorDbClient: Aerospike Vector Search Vector Client instance.
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Exit a context manager for the vector client.
        """
        self.close()
